/*----------------------------------------------------------------------
  File    : condep.c
  contents: measure strength of (conditional) dependences
  Author  : Christian Borgelt
  History : 1997.03.25 file created (as file cind.c)
            1998.02.02 option -u (unknown values characters) added
            1998.06.23 adapted to modified attset functions
            1999.04.18 simplified using the new module 'io'
            1999.04.30 use of temporary file removed
            2002.01.08 generalized to different evaluation measures
            2002.01.12 made independent of module ascons
            2002.01.22 option -w (weight distribution) added
            2003.01.17 normalization of possibilistic measures added
            2003.08.16 slight changes in error message output
            2006.02.16 reading a list of tests from domain file added
            2007.08.31 bug in tuple aggregation fixed (null values)
            2007.09.02 copying the attribute values simplified
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <math.h>
#include <assert.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef PT_PARSE
#define PT_PARSE
#endif
#include "ptree.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "condep"
#define DESCRIPTION "measure strength of (conditional) (in)dependences"
#define VERSION     "version 3.1 (2014.10.24)         " \
                    "(c) 1997-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error */
#define E_MEASURE  (-10)        /* unknown evaluation measure */
#define E_ATTCNT   (-11)        /* too few attributes */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_MEASURE -10 */  "unknown evaluation measure %s",
  /* E_ATTCNT  -11 */  "too few attributes (need at least 2)",
  /*           -12 */  "unknown error"
};

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- measure information --- */
  int  code;                    /* measure code */
  char *name;                   /* name of the measure */
} MINFO;                        /* (measure information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const MINFO m_prob[] = { /* --- probabilistic measures */
  { PT_INFGAIN, "infgain" },    /* information gain */
  { PT_INFSGR1, "infsgr1" },    /* sym. information gain ratio 1 */
  { PT_INFSGR2, "infsgr2" },    /* sym. information gain ratio 2 */
  { PT_QIGAIN,  "qigain"  },    /* quadratic information gain */
  { PT_QISGR1,  "qisgr1"  },    /* sym. quad. info. gain ratio 1 */
  { PT_QISGR2,  "qisgr2"  },    /* sym. quad. info. gain ratio 2 */
  { PT_GINISYM, "ginisym" },    /* symmetric Gini index */
  { PT_WDIFF,   "wdiff"   },    /* sum of weighted differences */
  { PT_CHI2,    "chi2"    },    /* chi^2 measure */
  { PT_CHI2NRM, "chi2nrm" },    /* normalized chi^2 measure */
  { -1,          NULL     }     /* sentinel */
};

static const MINFO m_poss[] = { /* --- possibilistic measures */
  { PT_SPCGAIN, "spcgain" },    /* specificity gain */
  { PT_SPCSGR1, "spcsgr1" },    /* sym. specificity gain ratio 1 */
  { PT_SPCSGR2, "spcsgr2" },    /* sym. specificity gain ratio 2 */
  { PT_WDIFF,   "wdiff"   },    /* weighted (absolute) difference */
  { PT_CHI2,    "chi2"    },    /* possibilistic chi^2 measure */
  { PT_MUTSPC,  "mutspc"  },    /* mutual specificity */
  { -1,          NULL     }     /* sentinel */
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan    = NULL; /* scanner (for domain file) */
static TABREAD *tread   = NULL; /* table reader */
static ATTSET  *attset  = NULL; /* attribute set */
static TABLE   *table   = NULL; /* table */
static ATTID   ptcnt    = 0;    /* number of probability trees */
static ATTID   ptsize   = 0;    /* size of probability tree array */
static PTREE   **ptrees = NULL; /* prob./poss. trees for testing */
static ATTID   *ids     = NULL; /* array of attribute identifiers */
static ATT     **atts   = NULL; /* array of needed attributes */
static FILE    *out     = NULL; /* output file */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/
#ifndef NDEBUG

static void del_ptrees (void)
{                               /* --- delete prob./poss. trees */
  if (!ptrees) return;          /* if there are no trees, abort */
  while (--ptcnt > 0)           /* delete all prob./poss. trees */
    if (ptrees[ptcnt]) pt_delete(ptrees[ptcnt], 0);
  free(ptrees);                 /* delete the tree vector */
}  /* del_ptrees() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (ptrees) del_ptrees(); \
  if (atts)   free(atts);   \
  if (ids)    free(ids);    \
  if (table)  tab_delete(table, 0); \
  if (attset) as_delete(attset);    \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on eval. measures */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate the startup message */
  printf("\nlist of evaluation measures (option -e#)\n");
  printf("measures for probabilistic mode\n");
  printf("  name       measure\n");
  for (i = 0; i < (int)(sizeof(m_prob)/sizeof(MINFO)-1); i++)
    printf("  %-9s  %s\n", m_prob[i].name,
           pt_mname(PT_PROB, m_prob[i].code));
  printf("\nmeasures for possibilistic mode\n");
  printf("  name       measure\n");
  for (i = 0; i < (int)(sizeof(m_poss)/sizeof(MINFO)-1); i++)
    printf("  %-9s  %s\n", m_poss[i].name,
           pt_mname(PT_POSS, m_poss[i].code));
  exit(0);                      /* abort porgram */
}  /* help() */

/*--------------------------------------------------------------------*/

static int code (const MINFO *tab, const char *name)
{                               /* --- get measure code */
  assert(tab && name);          /* check the function arguments */
  for ( ; tab->name; tab++)     /* look up name in table */
    if (strcmp(tab->name, name) == 0)
      return tab->code;         /* return the measure code */
  return -1;                    /* or an error indicator */
}  /* code() */

/*----------------------------------------------------------------------
ATTID next_test (ATTID cnt, ATTID cur, ATTID max, ATTID *ids)
cnt: number of attributes
cur: current number of conditions (-1 on first call)
max: maximum number of conditions
ids: vector of attribute identifiers,
     holds the two conditioned attributes followed by the conditions
----------------------------------------------------------------------*/

static ATTID next_test (ATTID cnt, ATTID cur, ATTID max, ATTID *ids)
{                               /* --- generate next dependence test */
  ATTID i, k;                   /* loop variable, buffer */
  ATTID *p;                     /* to traverse the identifiers */

  assert(ids                    /* check the function arguments */
  &&    (cnt >= 2) && (cur >= -1) && (cur <= max) && (max <= cnt-2));
  if (cur < 0) {                /* if not initialized, set first ids. */
    ids[0] = 0; ids[1] = 1; return 0; }
  for (k = cnt, p = ids+2 +(i = cur); --i >= 0; ) {
    (*--p)++;                   /* increase the condition identifier */
    if (*p  == ids[0]) (*p)++;  /* and make sure that it differs */
    if (*p  == ids[1]) (*p)++;  /* from the first two identifiers */
    if (--k == ids[1]) k--;     /* compute the upper bound */
    if (  k == ids[0]) k--;     /* for the current condition */
    if (*p <= k)                /* if the increased condition */
      break;                    /* does not exceed its upper bound, */
  }                             /* abort the loop */
  if (i < 0) {                  /* if no condition could be increased */
    if (++ids[1] >= cnt) {      /* try to increase the 2nd identifier */
      if (++ids[0] >= cnt-1) {  /* and on failure  the 1st identifier */
        if (++cur > max)        /* if the max. number of conditions */
          return -1;            /* is reached, abort the function, */
        ids[0] = 0;             /* otherwise reinit. and generate */
      }                         /* tests with one condition more */
      ids[1] = ids[0] +1;       /* reinitialize second identifier */
    }
    if (cur <= 0) return cur;   /* if no conditions to generate, */
    *p = i = 0;                 /* there is nothing else to do, */
  }                             /* otherwise init. first condition */
  while (1) {                   /* condition initialization loop */
    if (*p == ids[0]) (*p)++;   /* skip the first two identifiers */
    if (*p == ids[1]) (*p)++;   /* (must not appear as conditions) */
    if (++i >= cur) break;      /* if all conditions generated, abort */
    p[1] = *p+1; p++;           /* initialize the next condition */
  }
  return cur;                   /* return the number of conditions */
}  /* next_test() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domains file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *mname   = NULL;      /* name of evaluation measure */
  int     emode    = PT_PROB;   /* evaluation mode */
  int     measure  = 1;         /* evaluation measure */
  int     dwnull   = 0;         /* whether to distribute for nulls */
  int     mode     = AS_ATT|AS_NOXATT; /* table file read mode */
  double  param    = 2;         /* parameter for measure wdiff */
  ATTID   maxcon   = 0;         /* maximum number of conditions */
  ATTID   maxpt    = 32;        /* maximum number of trees */
  ATTID   concnt   = -1;        /* number of conditions */
  int     digits   = 6;         /* significant digits for output */
  PTREE   **pt;                 /* to traverse the prob./poss. trees */
  ATT     *att;                 /* to traverse the attributes */
  TUPLE   *tpl;                 /* to traverse the table */
  WEIGHT  wgt;                  /* buffer for instantiation weight */
  ATTID   m, c, x;              /* number of attributes/trees/conds. */
  TPLID   n, r;                 /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] domfile "
                     "[-d|-h hdrfile] tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-m       possibilistic mode (default: probabilistic)\n");
    printf("-e#      evaluation measure (default: infgain/spcgain)\n");
    printf("-!       print a list of evaluation measures\n");
    printf("-p#      parameter for evaluation measure wdiff\n");
    printf("-w       distribute tuple weight for unknown values\n");
    printf("-c#      maximum number of conditions           "
                    "(default: %"ATTID_FMT")\n", maxcon);
    printf("-k#      maximum number of parallel tests       "
                    "(default: %"ATTID_FMT")\n", maxpt);
    printf("-o#      significant digits for number output   "
                    "(default: %d)\n", digits);
    printf("-n       number of occurrences in last field\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("domfile  file containing domain descriptions "
                    "(in topological order)\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("resfile  file to write results to (optional)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                           break;
          case 'm': emode  = PT_POSS;                 break;
          case 'e': optarg = &mname;                  break;
          case 'p': param  =        strtod(s, &s);    break;
          case 'w': dwnull = 1;                       break;
          case 'c': maxcon = (ATTID)strtol(s, &s, 0); break;
          case 'k': maxpt  = (ATTID)strtol(s, &s, 0); break;
          case 'o': digits = (int)  strtol(s, &s, 0); break;
          case 'b': optarg = &blanks;                 break;
          case 'f': optarg = &fldseps;                break;
          case 'r': optarg = &recseps;                break;
          case 'u': optarg = &nullchs;                break;
          case 'C': optarg = &comment;                break;
          case 'n': mode  |= AS_WEIGHT;               break;
          case 'd': mode  |= AS_DFLT;                 break;
          case 'h': optarg = &fn_hdr;                 break;
          default : error(E_OPTION, *(--s));          break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 3) error(E_ARGCNT);  /* and the number of arguments */
  if ((fn_hdr) && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_dom || !*fn_dom) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  measure = (emode == PT_PROB)  /* get the evaluation measure code */
          ? ((mname) ? code(m_prob, mname) : PT_INFGAIN)
          : ((mname) ? code(m_poss, mname) : PT_SPCGAIN);
  if (measure < 0) error(E_MEASURE, mname);
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan));
  m = as_attcnt(attset);        /* parse domain descriptions */
  if (m < 2) error(E_ATTCNT);   /* check the number of attributes */

  /* --- read probabilitiy trees --- */
  while (scn_token(scan) != T_EOF) {
    if (ptcnt >= ptsize) {      /* if the tree vector is full */
      ptsize += (ptsize >= 32) ? ptsize >> 1 : 32;
      pt = (PTREE**)realloc(ptrees, (size_t)ptsize *sizeof(PTREE*));
      if (!pt) error(E_NOMEM);  /* enlarge the tree vector, */
      ptrees = pt;              /* copy the existing trees, */
    }                           /* and set the new vector */
    ptrees[ptcnt] = pt_parse(attset, scan, PT_CITEST|emode);
    if (!ptrees[ptcnt]) error(E_PARSE);
    ptcnt++;                    /* parse a description of a */
  }                             /* conditional independence test */
  ptsize = ptcnt;               /* note the number of tests */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)", m);
  if      (ptcnt > 1) fprintf(stderr, ", %"ATTID_FMT" queries", ptcnt);
  else if (ptcnt > 0) fprintf(stderr, ", 1 query");
  fprintf(stderr, "] done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  if (ptcnt > 0)                /* if trees have been read, */
    mode |= AS_NOXVAL;          /* the domains must be fixed */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  table = tab_create("table", attset, tpl_delete);
  if (!table) error(E_NOMEM);   /* create a data table */
  k = tab_read(table, tread, mode);
  if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(table);        /* get the number of attributes */
  n = tab_tplcnt(table);        /* and the number of data tuples */
  w = tab_tplwgt(table);        /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- reduce table --- */
  fprintf(stderr, "reducing table ... ");
  n = tab_reduce(table);        /* reduce the table for speed up */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- measure (conditional) (in)dependences --- */
  t = clock();                  /* start timer, open output file */
  if (fn_out && (strcmp(fn_out, "-") == 0)) fn_out = "";
  if (fn_out && *fn_out) { out = fopen(fn_out, "w"); }
  else                   { out = stdout; fn_out = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_out);
  if (!out) error(E_FOPEN, fn_out);
  if (ptsize <= 0) {            /* if no cond. indep. tests are given */
    ids    = (ATTID*) malloc((size_t)m *sizeof(ATTID));
    ptrees = (PTREE**)calloc((size_t)maxpt, sizeof(PTREE*));
    if (!ids || !ptrees) error(E_NOMEM);
    if (maxcon > m-2)           /* create id and tree vectors and */
      maxcon = m-2;             /* limit the number of conditions */
  }                             /* (two attributes are conditioned) */
  atts = (ATT**)malloc((size_t)m *sizeof(ATT*));
  if (!atts) error(E_NOMEM);    /* create an attribute vector */
  for (pt = ptrees; 1; ) {      /* conditional independence test loop */
    memset(atts, 0, (size_t)m *sizeof(ATT*));
    if (ptsize > 0) {           /* if a list of tests is given */
      ptcnt = (ptsize > maxpt)  /* get the number of trees that are */
            ? maxpt : ptsize;   /* to be processed in the next run */
      for (x = 0; x < ptcnt; x++) { /* traverse prob./poss. trees */
        att = pt_att(pt[x]);    /* note the conditioned attribute */
        atts[att_id(att)] = att;/* of each prob./poss. tree */
        for (c = pt_concnt(pt[x]); --c >= 0; ) {
          att = pt_con(pt[x], c);
          atts[att_id(att)] = att;
        }                       /* note all condition attributes */
      } }                       /* of the prob./poss. tree */
    else {                      /* if to generate cond. indep. tests */
      for (ptcnt = 0; ptcnt < maxpt; ptcnt++) {
        concnt = next_test(m, concnt, maxcon, ids);
        if (concnt < 0) break;  /* generate the next dependence test */
        atts[ids[0]] = att = as_att(attset, ids[0]);
        pt[ptcnt] = pt_create(att, emode);
        if (!pt[ptcnt]) error(E_NOMEM); /* create a prob./poss. tree */
        atts[ids[1]] = att = as_att(attset, ids[1]);
        if (pt_conadd(pt[ptcnt], 0, att) < 0)
          error(E_NOMEM);       /* add 2nd attribute as 1st condition */
        for (c = 0; c < concnt; c++) {
          x = ids[2+c];         /* traverse the remaining identifiers */
          atts[x] = att = as_att(attset, x);
          if (pt_conadd(pt[ptcnt], c, att) < 0)
            error(E_NOMEM);     /* add the conditions according to */
        }                       /* the generated identifier vector */
      }                         /* and note the used attributes */
    }                           /* (for copying the attribute values) */
    if (ptcnt <= 0) break;      /* if no test is left, abort the loop */
    for (x = c = 0; x < m; x++) /* collect the occurring attributes */
      if (atts[x]) atts[c++] = atts[x];
    for (r = tab_tplcnt(table); --r >= 0; ) {
      tpl = tab_tpl(table, r);  /* traverse the tuples and */
      for (x = c; --x >= 0; )   /* copy the needed attribute values */
        *att_inst(atts[x]) = *tpl_colval(tpl, att_id(atts[x]));
      wgt = tpl_getwgt(tpl);    /* get the tuple weight */
      for (x = ptcnt; --x >= 0;)/* traverse the prob./poss. trees */
        if (pt_aggr(pt[x], wgt, dwnull) < 0)
          error(E_NOMEM);       /* aggregate the current tuple */
    }                           /* for all trees to be processed */
    for (x = 0; x < ptcnt; x++){/* traverse the prob./poss. trees */
      c = pt_concnt(pt[x]);     /* print the two conditioned atts. */
      fprintf(out, "I(%s", att_name(pt_att(pt[x])));
      fprintf(out, ", %s", att_name(pt_con(pt[x], --c)));
      if (--c >= 0) fprintf(out, " | %s", att_name(pt_con(pt[x], c)));
      while (--c >= 0)          /* print the list of conditions */
        fprintf(out, ", %s", att_name(pt_con(pt[x], c)));
      fprintf(out, ") = ");     /* print the evaluation result */
      fprintf(out, "%.*g", digits, pt_depend(pt[x], measure, &param));
      fprintf(out, ";\n");      /* terminate the output line */
      pt_delete(pt[x], 0); pt[x] = NULL;
    }                           /* delete the tree and clean up */
    if (ptsize > 0) {           /* if cond. indep. tests are given */
      pt     += ptcnt;          /* remove the processed trees */
      ptsize -= ptcnt;          /* from the list of tests */
      if (ptsize <= 0) break; } /* if all tests have been processed */
    else if (concnt < 0)        /* or all tests have been generated, */
      break;                    /* abort the test loop */
  }
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_out);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "done.\n");   /* print a success message */

  /* --- clean up --- */
  #ifndef NDEBUG
  free(ptrees);                 /* delete the tree vector, */
  free(ids);                    /* the identifier vector, */
  tab_delete(table, 0);         /* the table, and */
  as_delete(attset);            /* the attribute set */
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  return 0;                     /* return 'ok' */
}  /* main() */
