/*----------------------------------------------------------------------
  File    : gmutil.c
  Contents: utility functions for inducing graphical models
  Author  : Christian Borgelt
  History : 2002.03.03 file created as gramod2.c
            2006.03.04 utility functions moved to separate file
            2006.03.10 condition copying in gmu_eval only if dep >= 0
            2006.03.12 most fprintf replaced by fputs and fputc
            2007.01.29 bug in function gmu_ptaggr fixed (unknown value)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "gmutil.h"
#include "arrays.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- connected component --- */
  ATTID id;                     /* identifier of representative */
  ATTID cnt;                    /* number of represented nodes */
} COMP;                         /* (connected component) */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

int gmu_init (GMID *gmid, ATTID cnt, int half)
{                               /* --- create an edge table */
  ATTID  i, k, s;               /* loop variables, table size */
  GMEDGE **tab, *e;             /* created edge table */

  assert(gmid);                 /* check the function argument */
  gmid->adjs   = NULL;          /* clear working buffer variable */
  gmid->attcnt = cnt;           /* store the number of attributes */
  if (half) { s = cnt*(cnt-1)/2; if (half > 0) s += cnt; }
  else      { s = cnt*cnt; }    /* compute the table size */
  gmid->etab = tab = (GMEDGE**)malloc((size_t)(cnt+s) *sizeof(GMEDGE*));
  if (!tab) return -1;          /* create the vector of columns */
  e = tab[0] = (GMEDGE*)malloc((size_t)s *sizeof(GMEDGE));
  if (!e) { gmid->etab = NULL; free(tab); return -1; }
  gmid->etab  = tab;            /* create the table entries */
  gmid->edges = tab +cnt;       /* and set the edge vector */
  half = (half > 0) ? 1 : ((half < 0) ? 0 : -1);
  for (e += s, i = cnt; --i >= 0; ) {
    for (k = (half >= 0) ? i +half : cnt; --k >= 0; ) {
      --e; e->y = k; e->x = i; e->mark = 0; e->eval = 0.0; }
    tab[i] = e;                 /* initialize the edge variables and */
  }                             /* organize the edges into a table */
  return 0;                     /* return 'ok' */
}  /* gmu_init() */

/*--------------------------------------------------------------------*/

ATTID gmu_edges (GMID *gmid)
{                               /* --- collect selectable edges */
  ATTID  i, k, n, c;            /* loop variables, condition id */
  GMEDGE **tab, **p, *e;        /* created edge table */

  assert(gmid);                 /* check the function argument */
  tab = gmid->etab;             /* get the edge table and */
  n   = gmid->attcnt;           /* the number of attributes */
  for (i = n; --i >= 0; ) {     /* traverse the attributes */
    for (k = gm_cddcnt(gmid->gm, i); --k >= 0; ) {
      c = gm_cddid(gmid->gm, i, k);
      if (i > c) tab[i][c].mark = 1;
      else       tab[c][i].mark = 1;
    }                           /* traverse the candidate attributes */
  }                             /* and mark the selectable edges */
  for (i = n; --i >= 0; ) {     /* traverse the attributes */
    for (k = gm_concnt(gmid->gm, i); --k >= 0; ) {
      c = gm_conid(gmid->gm, i, k);
      if (i > c) tab[i][c].mark = 2;
      else       tab[c][i].mark = 2;
    }                           /* traverse the existing conditions */
  }                             /* and mark the mandatory edges */
  gm_clear(gmid->gm, -1);       /* clear the graphical model */
  p = gmid->edges;              /* get the edge collection vector */
  for (i = n; --i > 0; ) {      /* and traverse the attributes */
    for (e = tab[i] +(k = i); --k >= 0; )
      if ((--e)->mark > 0) *p++ = e;
  }                             /* collect the selectable edges */
  gmid->selcnt = 0;             /* store the numbers of edges */
  gmid->potcnt = gmid->inicnt = (ATTID)(p -gmid->edges);
  return gmid->potcnt;          /* return number of potential edges */
}  /* gmu_edges() */

/*--------------------------------------------------------------------*/

int gmu_eval (GMID *gmid, ATTID cnt, int dep)
{                               /* --- evaluate selected edges */
  ATTID  i, k, n;               /* loop variables, number of trees */
  GMEDGE **edges;               /* edges to evaluate */
  PTREE  **trees, *pt;          /* to traverse the prob./poss. trees */
  ATT    *att;                  /* to traverse the attributes */
  double eval;                  /* edge evaluation */

  assert(gmid);                 /* check the function argument */
  trees = (PTREE**)calloc((size_t)gmid->ptmax, sizeof(PTREE*));
  if (!trees) return -1;        /* create a vector for the trees */
  edges = gmid->edges;          /* get the edge vector and */
  if (cnt < 0) cnt = gmid->potcnt;  /* the number of edges */
  while (cnt > 0) {             /* while there are edges left */

    /* -- create probability/possibility trees -- */
    n = (cnt > gmid->ptmax) ? gmid->ptmax : cnt;
    for (i = n; --i >= 0; ) {   /* traverse the edges to process */
      att = tab_col(gmid->table, edges[i]->x);
      trees[i] = pt = pt_create(att, gmid->gm->type);
      if (!pt)                  /* create a prob./poss. tree */
        return gmu_ptclean(trees, n, NULL);
      if ((dep >= 0)            /* copy existing conditions */
      &&  (pt_concopy(pt, gm_ptree(gmid->gm, edges[i]->x)) != 0))
        return gmu_ptclean(trees, n, NULL);
      if (gmid->gm->tplcnt > 0) /* set the normalization factor */
        pt_setnorm(pt, (float)gmid->gm->tplcnt);
      k = edges[i]->y;          /* skip loops (pure evaluation) */
      if (k == edges[i]->x) continue;
      if (pt_conadd(pt, pt_concnt(pt), tab_col(gmid->table, k)) != 0)
        return gmu_ptclean(trees, n, NULL);
    }                           /* add the source as last condition */

    /* -- evaluate probability/possibility trees -- */
    if (gmu_ptaggr(trees, n, gmid->table) != 0)
      return gmu_ptclean(trees, n, NULL);
    for (i = n; --i >= 0; ) {   /* aggregate the tuples and then */
      pt = trees[i];            /* traverse and evaluate the trees */
      k  = pt_concnt(pt);       /* get the number of conditions */
      if (dep) {                /* if conditional dependence test */
        eval = pt_depend(pt, gmid->measure, gmid->params);
        if (isnan(eval))        /* compute cond. dependence measure */
          return gmu_ptclean(trees, n, NULL);
        if (gmid->log) {        /* if a log file is given */
          fputs(att_name(pt_att(pt)),      gmid->log);
          fputc(' ',                       gmid->log);
          fputs(att_name(pt_con(pt, --k)), gmid->log);
        } }                     /* write evaluated attribute pair */
      else {                    /* if dependence strength measurement */
        eval = pt_eval(pt, gmid->measure, gmid->params);
        if (!isnan(eval) && (gmid->lwise >= 0))
          eval = pt_local(pt, gmid->lwise, gmid->mrgimp);
        if ( isnan(eval))       /* do local structure learning */
          return gmu_ptclean(trees, n, NULL);
        if (gmid->log)          /* if a log file is given */
          fputs(att_name(pt_att(pt)), gmid->log);
      }                         /* write conditioned attribute */
      if (gmid->log) {          /* if a log file is given */
        if (k > 0) fputs(" |", gmid->log);
        while (--k >= 0) {      /* while there are conditions */
          fputc(' ',                     gmid->log);
          fputs(att_name(pt_con(pt, k)), gmid->log);
        }                       /* list the conditions */
        fprintf(gmid->log, "  : %g\n", eval);
      }                         /* write the tree evaluation */
      edges[i]->eval = eval;    /* note the evaluation result */
      pt_delete(pt, 0); trees[i] = NULL;
    }                           /* delete the prob./poss. tree */
    cnt -= n; edges += n;       /* skip the processed edges */
  }
  free(trees);                  /* delete the tree vector */
  return 0;                     /* return 'ok' */
}  /* gmu_eval() */

/*--------------------------------------------------------------------*/

ATTID gmu_filter (GMID *gmid)
{                               /* --- filter evaluated edges */
  ATTID  n, k = 0;              /* loop variable */
  GMEDGE *e, **p, **q;          /* to traverse the edges */

  assert(gmid);                 /* check the function argument */
  p = q = gmid->edges;          /* traverse the edges in the vector */
  for (n = gmid->potcnt; --n >= 0; q++) {
    if ( ((*q)->mark == 2)      /* if the edge is mandatory or */
    ||  (((*q)->mark == 1)      /* has minimum evaluation, keep it */
    &&   ((*q)->eval >= gmid->thresh))) {
      e = *p; *p++ = *q; *q = e; k++; }
  }                             /* collect qualifying edges */
  return gmid->potcnt = k;      /* return the number of edges */
}  /* gmu_filter() */

/*--------------------------------------------------------------------*/

ATTID gmu_span (GMID *gmid)
{                               /* --- find edges of spanning tree */
  ATTID  i, k, n;               /* loop variable, counter */
  ATTID  x, y;                  /* node identifiers */
  GMEDGE *e, **p, **q;          /* to traverse the edges */
  COMP   *c;                    /* vector of connected components */

  assert(gmid);                 /* check the function arguments */
  c = (COMP*)malloc((size_t)gmid->attcnt *sizeof(COMP));
  if (!c) return -1;            /* create a connected comp. vector */
  for (c += i = gmid->attcnt; --i >= 0; ) {
    (--c)->id = i; c->cnt = 1;} /* init. the connected components */
  ptr_qsort(gmid->edges, (size_t)(n = gmid->potcnt), +1, gmu_cmp, NULL);
  p = q = gmid->edges;          /* sort edges according to weight, */
  for (k = 0; --n >= 0; q++) {  /* then traverse them descendingly */
    x = (*q)->x;                /* find first  node's component id. */
    while (x != c[x].id) x = c[x].id;
    y = (*q)->y;                /* find second node's component id. */
    while (y != c[y].id) y = c[y].id;
    if (x == y)                 /* if the components are the same, */
      continue;                 /* skip the edge (avoid loops) */
    if (c[x].cnt < c[y].cnt) { c[y].cnt += c[x].cnt; c[x].id = y; }
    else                     { c[x].cnt += c[y].cnt; c[y].id = x; }
    e = *p; *p++ = *q; *q = e;  /* combine the connected components */
    if (++k >= gmid->attcnt-1)  /* and collect the selected edges */
      break;                    /* if edges form a spanning tree, */
  }                             /* the function can be aborted early */
  free(c);                      /* delete connected comp. vector, */
  return gmid->selcnt = k;      /* return number of selected edges */
}  /* gmu_span() */

/*--------------------------------------------------------------------*/

void gmu_cleanup (GMID *gmid)
{                               /* --- clean up induction variables */
  if (gmid->etab) { free(gmid->etab[0]); free(gmid->etab); }
  if (gmid->adjs) free(gmid->adjs);
}  /* gmu_cleanup() */

/*--------------------------------------------------------------------*/

int gmu_cmp (const void *p1, const void *p2, void *data)
{                               /* --- compare two edges */
  if (((GMEDGE*)p1)->mark > ((GMEDGE*)p2)->mark) return -1;
  if (((GMEDGE*)p1)->mark < ((GMEDGE*)p2)->mark) return +1;
  if (((GMEDGE*)p1)->eval > ((GMEDGE*)p2)->eval) return -1;
  if (((GMEDGE*)p1)->eval < ((GMEDGE*)p2)->eval) return +1;
  return 0;                     /* return sign of eval. difference */
}  /* gmu_cmp() */

/*--------------------------------------------------------------------*/

int gmu_ptaggr (PTREE **trees, ATTID n, TABLE *table)
{                               /* --- aggregate tuples for trees */
  ATTID i, k;                   /* loop variables */
  ATTID acnt, used;             /* number of (used) attributes */
  ATT   **atts, *a;             /* attributes used in the trees */
  TUPLE *tpl;                   /* to traverse the tuples */
  float wgt;                    /* instantiation weight */

  assert(trees && table && (n > 0)); /* check function arguments */
  acnt = tab_colcnt(table);     /* get the number of attributes */
  atts = (ATT**)calloc((size_t)acnt, sizeof(ATT*));
  if (!atts) return -1;         /* create an attribute vector */
  for (k = n; --k >= 0; ) {     /* traverse the probability trees */
    a = pt_att(trees[k]);       /* get the conditioned attribute */
    atts[att_id(a)] = a;        /* and note it in the vector */
    for (i = pt_concnt(trees[k]); --i >= 0; ) {
      a = pt_con(trees[k], i); atts[att_id(a)] = a; }
  }                             /* collect all condition attributes */
  for (used = i = 0; i < acnt; i++)
    if (atts[i] != NULL)        /* remove gaps in the attribute list */
      atts[used++] = atts[i];   /* and count the used attributes */
  if (used >= acnt/3) {         /* if many attributes are used */
    free(atts);                 /* delete the attribute vector */
    for (i = tab_tplcnt(table); --i >= 0; ) {
      tpl = tab_tpl(table, i);  /* traverse the tuples and */
      tpl_toas(tpl);            /* copy them to the attribute set */
      wgt = tpl_getwgt(tpl);    /* get the weight of the tuple */
      for (k = n; --k >= 0; )   /* traverse the prob./poss. trees */
        if (pt_aggr(trees[k], wgt, 0) < 0) {
          free(atts); return -1; }
    } }                         /* aggregate the tuples for all trees */
  else {                        /* if only few attributes are needed */
    for (i = tab_tplcnt(table); --i >= 0; ) {
      tpl = tab_tpl(table, i);  /* traverse the tuples and */
      for (k = used; --k >= 0;) /* set the needed attribute values */
        *att_inst(atts[k]) = *tpl_colval(tpl, att_id(atts[k]));
      wgt = tpl_getwgt(tpl);    /* get the weight of the tuple */
      for (k = n; --k >= 0; )   /* traverse the prob./poss. trees */
        if (pt_aggr(trees[k], wgt, 0) < 0) {
          free(atts); return -1; }
    }                           /* aggregate the tuples for all trees */
    free(atts);                 /* delete the attribute vector */
  }
  return 0;                     /* return 'ok' */
}  /* gmu_ptaggr() */

/*--------------------------------------------------------------------*/

int gmu_ptclean (PTREE **trees, ATTID n, void *p)
{                               /* --- clean up prob./poss. trees */
  while (--n >= 0)              /* delete all existing trees */
    if (trees[n]) pt_delete(trees[n], 0);
  if (trees) free(trees);       /* delete the tree vector */
  if (p)     free(p);           /* and the additional vector */
  return -1;                    /* return 'error' */
}  /* gmu_ptclean() */
