/*----------------------------------------------------------------------
  File    : gramod3.c
  Contents: graphical model management (based on directed graphs)
            spanning tree construction and greedy condition selection
  Author  : Christian Borgelt
  History : 1995.11.03 file created as ines.c
            1995.11.06 modified K2 programmed
            1996.03.05 lower bound for improvement added
            1996.03.06 adapted to modified set of evaluation measures
            1997.03.07 optimum weight spanning tree method added
            1996.03.26 use of minimal improvement corrected
            1996.06.18 bug in use of minimal improvement fixed
            1999.04.27 redesign of network induction completed
            1999.12.18 bug in owst() fixed (loop variable)
            2001.12.14 optimum spanning tree extension added
            2002.03.03 network induction functions moved out of ines.c
            2002.07.04 local structure learning (re)added
            2006.03.02 aggregation of tuples for trees optimized
            2006.03.03 cond. independence tests construction added
            2006.03.12 function extst() redesigned (half edge table)
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "gmutil.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Optimum Weight Spanning Tree Construction
----------------------------------------------------------------------*/

static int owst (GMID *tid)
{                               /* --- optimum weight spanning tree */
  ATTID  i, n;                  /* loop variables */
  ATTID  x, y;                  /* node identifiers */
  GMEDGE **p, **q;              /* to traverse the edges */
  ATTID  *cnts;                 /* number of incident edges per node */

  assert(tid);                  /* check the function argument */

  /* --- find the tree edges --- */
  if ((gmu_edges (tid) < 0)     /* collect the selectable edges */
  ||  (gmu_eval  (tid, -1, -1) < 0)
  ||  (gmu_filter(tid) < 0)     /* evaluate and filter the edges */
  ||  (gmu_span  (tid) < 0))    /* and construct a spanning tree */
    return -1;                  /* from the remaining edges */

  /* --- direct the tree edges --- */
  cnts = (ATTID*)malloc((size_t)tid->attcnt *sizeof(ATTID));
  if (!cnts) return -1;         /* create an edge counter vector */
  for (cnts += i = tid->attcnt; --i >= 0; )
    *--cnts = 0;                /* initialize the edge counters */
  for (p = tid->edges +(i = tid->selcnt); --i >= 0; ) {
    cnts[(*--p)->x]++;          /* traverse the edges and */
    cnts[(*  p)->y]++;          /* determine for each node */
  }                             /* the number of incident edges */
  gm_conrem(tid->gm, -1, 0);    /* remove all conditions */
  for (n = tid->selcnt; n > 0;) { /* while there are undir. edges */
    for (p = q = tid->edges; --n >= 0; q++) {
      x = (*q)->x;              /* traverse the edges and */
      y = (*q)->y;              /* get the connected nodes */
      if      (cnts[x] <= 1) {  /* if edge can be dir. to 1st node */
        if (gm_conadd(tid->gm, x, y) < 0) {
          free(cnts); return -1; } }   /* make 2nd node the parent */
      else if (cnts[y] <= 1) {  /* if edge can be dir. to 2nd node */
        if (gm_conadd(tid->gm, y, x) < 0) {
          free(cnts); return -1; } }   /* make 1st node the parent */
      else {                    /* if the edge cannot be directed, */
        *p++ = *q; continue; }  /* keep it for the next traversal */
      cnts[x]--; cnts[y]--;     /* remove the edge from the counters */
    }                           /* (counters refer to undir. edges) */
    n = (ATTID)(p -tid->edges); /* get the number of remaining edges */
  }                             /* (that have not been directed yet) */
  free(cnts);                   /* delete the counter vector */
  return 0;                     /* return 'ok' */
}  /* owst() */

/*--------------------------------------------------------------------*/

static int extst (GMID *tid)
{                               /* --- extension of a spanning tree */
  ATTID  i, k, n;               /* loop variables, number of edges */
  ATTID  x, y;                  /* node identifiers */
  GMEDGE *e, **p, **q;          /* to traverse the edges */

  assert(tid);                  /* check the function argument */

  /* --- collect possible edges --- */
  p = tid->edges;               /* traverse the attributes */
  for (i = tid->attcnt; --i >= 0; ) {
    n = gm_depcnt(tid->gm, i);  /* get the number of children */
    if (n <= 0) continue;       /* skip childless nodes (leaves) */
    if (gm_concnt(tid->gm, i) > 0) {  /* if there is a parent */
      y = gm_conid(tid->gm, i, 0);    /* (this is not the root) */
      for (k = n; --k >= 0; ) { /* traverse the children */
        x = gm_depid(tid->gm, i, k);
        e = (x > y) ? tid->etab[x] +y : tid->etab[y] +x;
        if (e->mark <= 0) continue;
        if (x > y) e->dir = -1;
        else     { e->dir = +1; e->x = y; e->y = x; }
        *p++ = e;               /* collect edges from the parent */
      }                         /* (these edges have a fixed dir., */
    }                           /* namely from parent to child) */
    while (--n > 0) {           /* traverse the children */
      y = gm_depid(tid->gm, i, n);
      for (k = n; --k >= 0; ) { /* traverse the preceding children */
        x = gm_depid(tid->gm, i, k);
        e = (x > y) ? tid->etab[x] +y : tid->etab[y] +x;
        if (e->mark <= 0) continue;
        e->dir = 0; *p++ = e;   /* collect edges between children */
      }                         /* (these do not have a fixed dir., */
    }                           /* but can be oriented either way) */
  }
  n = (ATTID)(p -tid->edges);   /* compute the number of edges */

  /* --- build tree extension --- */
  tid->potcnt = n;              /* set number of selectable edges */
  if ((gmu_eval  (tid, n, 1) < 0)
  ||  (gmu_filter(tid) < 0)     /* evaluate and filter the edges */
  ||  (gmu_span  (tid) < 0))    /* and select them in such a way */
    return -1;                  /* that they do not form a cycle */

  /* --- direct the extension edges --- */
  p = q = tid->edges;           /* traverse the selected edges */
  for (n = tid->selcnt; --n >= 0; q++) {
    if      ((*q)->dir == -1)
      gm_conadd(tid->gm, (*q)->x, (*q)->y);
    else if ((*q)->dir == +1)
      gm_conadd(tid->gm, (*q)->y, (*q)->x);
    else *p++ = *q;             /* direct edges with fixed direction */
  }                             /* (from parent to child) */
  for (n = (ATTID)(p -tid->edges); n > 0; ) {
    k = n;                      /* while there are undirected edges */
    for (p = q = tid->edges; --n >= 0; q++) {
      if      (gm_concnt(tid->gm, (*q)->x) > 1)
        gm_conadd(tid->gm, (*q)->y, (*q)->x);
      else if (gm_concnt(tid->gm, (*p)->y) > 1)
        gm_conadd(tid->gm, (*q)->x, (*q)->y);
      else *p++ = *q;           /* an edge can be directed */
    }                           /* if one node has two parents */
    n = (ATTID)(p -tid->edges); /* get the number of remaining edges */
    if (k <= n) { n--; p--;     /* if no edge has been directed */
      gm_conadd(tid->gm, (*p)->x, (*p)->y); }
  }                             /* direct an arbitrary edge */
  return 0;                     /* return 'ok' */
}  /* extst() */

/*--------------------------------------------------------------------*/

double gm_tree (GRAMOD *gm, TABLE *table, int search, int measure,
                double *params, double minval, ATTID maxtest, FILE *log)
{                               /* --- (extended) tree construction */
  double tplcnt;                /* number of tuples (weight) */
  GMID   tid;                   /* structure of induction variables */
  int    err;                   /* whether an error occurred */

  assert(gm && table);          /* check the function arguments */
  if ((search < GM_OWST) || (search > GM_EXTST))
    return NAN;                 /* check the search identifier */
  if (gm->attcnt <= 0)          /* if there are no attributes, */
    return 0.0;                 /* simply abort the function */
  tplcnt = tab_getwgt(table, 0, TPLID_MAX);
  if ((gm->type & GM_POSS)      /* for poss. graphical models */
  &&  (tab_opc(table, TAB_COND) != 0))
    return NAN;                 /* compute one point coverages */
  gm_setnorm(gm, tplcnt);       /* set number of tuples for norm. */
  tid.gm        = gm;           /* initialize the */
  tid.table     = table;        /* induction variables */
  tid.search    = search;
  tid.measure   = measure;
  tid.params[0] = params[0];
  tid.params[1] = params[1];
  tid.thresh    = minval;
  tid.mrgimp    =  0;
  tid.lwise     = -1;
  tid.maxcon    =  0;
  tid.ptmax     = (maxtest <= 0) ? 1 : maxtest;
  tid.log       = log;
  if (gmu_init(&tid, gm->attcnt, -1) < 0)
    return NAN;                 /* init. edge table (half, no diag.) */
  err = owst(&tid);             /* induce a spanning tree */
  if ((err == 0) && (search == GM_EXTST))
    err = extst(&tid);          /* extend the spanning tree */
  gmu_cleanup(&tid);            /* clean up the induction variables */
  if (err) return NAN;          /* and check for an error */
  gm_clear(gm, 1);              /* clear the graphical model */
  if ((gm_estim(gm, table, 0) != 0)
  ||  isnan(gm_eval(gm, measure, params)))
    return NAN;                 /* rebuild the graphical model */
  return gm->eval;              /* return the evaluation result */
}  /* gm_tree() */

/*----------------------------------------------------------------------
  Greedy Condition Selection
----------------------------------------------------------------------*/

static int consel (GMID *gmid)
{                               /* --- condition selection */
  ATTID  i, k, n;               /* loop variables, number of edges */
  GMEDGE *e, **p, *best;        /* to traverse the edges, best edge */
  ATTID  y;                     /* node identifier */
  int    top;                   /* topological order flag */

  assert(gmid);                 /* check the function argument */
  for (i = gmid->attcnt; --i >= 0; )
    gmid->edges[i] = gmid->etab[i] +i;
  if (gmu_eval(gmid, gmid->attcnt, 0) != 0)
    return -1;                  /* compute the initial evaluation */

  /* --- mark possible edges --- */
  top = (gmid->search == GM_TOPORD);
  for (i = gmid->attcnt; --i >= 0; ) {
    for (k = gm_cddcnt(gmid->gm, i); --k >= 0; ) {
      y = gm_cddid(gmid->gm, i, k);
      if (!top || (y < i)) gmid->etab[i][y].mark = 1;
    }                           /* traverse the candidate attributes */
    for (k = gm_concnt(gmid->gm, i); --k >= 0; ) {
      y = gm_conid(gmid->gm, i, k);
      if (!top || (y < i)) gmid->etab[i][y].mark = 0;
    }                           /* traverse the existing conditions */
  }                             /* and mark the selectable edges */

  /* --- select conditions --- */
  while (1) {                   /* condition selection loop */
    p = gmid->edges;            /* traverse the edge table */
    for (i = gmid->attcnt; --i >= 0; ) {
      e = gmid->etab[i];        /* get the next table column */
      if (e[i].mark < 0)        /* if an attribute is marked, */
        continue;               /* no conditions can be added */
      if (gm_concnt(gmid->gm, i) >= gmid->maxcon) {
        e[i].mark = -1;         /* if the maximum number is reached, */
        continue;               /* no more conditions can be added, */
      }                         /* so mark this attribute */
      for (n = 0, e += k = (top ? i : gmid->attcnt); --k >= 0; )
        if ((--e)->mark > 0) { *p++ = e; n++; }
      if (n <= 0) e[i].mark =-1;/* collect the selectable edges and */
    }                           /* mark attributes without edges */
    n = (ATTID)(p-gmid->edges); /* compute the number of edges */
    if (n <= 0) return 0;       /* if there are no edges left, abort */
    if (gmu_eval(gmid, n, 0) != 0)
      return -1;                /* evaluate the selected edges */
    for (i = gmid->attcnt; --i >= 0; ) {
      e = gmid->etab[i];        /* get the next table column */
      if (e[i].mark < 0)        /* for marked attributes */
        continue;               /* no conditions were tested */
      best = e +i;              /* traverse the selectable conditions */
      for (e += k = (top ? i : gmid->attcnt); --k >= 0; ) {
        if (((--e)->mark > 0) && (e->eval > best->eval))
          best = e;             /* traverse the edges and */
      }                         /* find the best condition */
      if ((best->y == i)
      ||  (best->eval -e[i].eval < gmid->thresh)) {
        e[i].mark = -1;         /* if adding the condition does not */
        continue;               /* lead to a sufficient improvement, */
      }                         /* stop selecting conditions */
      if (gm_conadd(gmid->gm, i, best->y) < 0)
        return -1;              /* add the selected condition */
      best->mark = 0;           /* and invalidate (unmark) it */
      e[i].eval  = best->eval;  /* note the new base evaluation */
    }                           /* (relative to which the next */
  }                             /* improvement will be measured) */
  return 0;                     /* return 'ok' */
}  /* consel() */

/*--------------------------------------------------------------------*/

static int noloop (GMID *gmid)
{                               /* --- selection avoiding loops */
  ATTID  i, k, n;               /* loop variables, number of edges */
  GMEDGE *e, **p, *best = NULL; /* to traverse the edges, best edge */
  double base, imp;             /* evaluation and improvement */

  assert(gmid);                 /* check the function argument */
  for (i = gmid->attcnt; --i >= 0; )
    gmid->edges[i] = gmid->etab[i] +i;
  if (gmu_eval(gmid, gmid->attcnt, 0) != 0)
    return -1;                  /* compute the initial evaluation */

  /* --- mark possible edges --- */
  for (i = gmid->attcnt; --i >= 0; ) {
    for (k = gm_cddcnt(gmid->gm, i); --k >= 0; )
      gmid->etab[i][gm_cddid(gmid->gm, i, k)].mark = 1;
    for (k = gm_concnt(gmid->gm, i); --k >= 0; )
      gmid->etab[i][gm_conid(gmid->gm, i, k)].mark = 0;
  }                             /* mark the selectable edges */

  /* --- select conditions --- */
  while (1) {                   /* condition selection loop */
    p = gmid->edges;            /* traverse the edge table */
    for (i = gmid->attcnt; --i >= 0; ) {
      e = gmid->etab[i];        /* get the next table column */
      if (e[i].mark < 0)        /* if an attribute is marked, */
        continue;               /* no conditions can be added */
      if (gm_concnt(gmid->gm, i) >= gmid->maxcon) {
        e[i].mark = -1;         /* if the maximum number is reached, */
        continue;               /* no more conditions can be added, */
      }                         /* so mark this attribute */
      for (n = 0, e += k = gmid->attcnt; --k >= 0; ) {
        if ((--e)->mark <= 0)   /* traverse the edges/conditions, */
          continue;             /* but skip unmarked edges */
        if (gm_check(gmid->gm, i, k) != 0)
          e->mark = 0;          /* remove edges that create a loop */
        else {                  /* and collect the selectable edges */
          if (!best || (i == best->x)) *p++ = e;
          n++;                  /* (collect edges only for the */
        }                       /* child of the previous step) */
      }
      if (!n) e[i].mark = -1;   /* mark attributes for which */
    }                           /* no edges (conditions) are left */
    n = (ATTID)(p-gmid->edges); /* compute the number of edges */
    if (gmu_eval(gmid, n, 0) != 0)
      return -1;                /* evaluate the selected edges */
    imp = -INFINITY;            /* traverse the table columns */
    for (i = gmid->attcnt; --i >= 0; ) {
      e = gmid->etab[i];        /* get the next table column */
      if (e[i].mark < 0)        /* for a marked attribute */
        continue;               /* no conditions were tested */
      base = e[i].eval;         /* get the base evaluation */
      for (e += k = gmid->attcnt; --k >= 0; ) {
        if (((--e)->mark > 0) && (e->eval -base > imp)) {
          imp = e->eval -base; best = e; }
      }                         /* traverse the edges and */
    }                           /* find the best condition */
    if (imp < gmid->thresh)     /* if the improvement is too small, */
      return 0;                 /* abort the search */
    if (gm_conadd(gmid->gm, best->x, best->y) < 0)
      return -1;                /* add the selected condition */
    best->mark = 0;             /* and invalidate (unmark) it */
    e = gmid->etab[best->x] +best->x;
    e->eval = best->eval;       /* note the new base evaluation */
  }
  return 0;                     /* return 'ok' */
}  /* noloop() */

/*--------------------------------------------------------------------*/

double gm_consel (GRAMOD *gm, TABLE *table, int search, int measure,
                  double *params, double minimp, int lwise,
                  double mrgimp, ATTID maxcon, ATTID maxtest, FILE *log)
{                               /* --- induce by selecting conditions */
  double tplcnt;                /* number of tuples (weight) */
  GMID   gmid;                  /* structure of induction variables */
  int    err;                   /* whether an error occurred */

  assert(gm && table);          /* check the function arguments */
  if ((search < GM_TOPORD) || (search > GM_UNRES))
    return NAN;                 /* check the method identifier */
  if (gm->attcnt <= 0)          /* if there are no attributes, */
    return 0.0;                 /* simply abort the function */
  tplcnt = tab_getwgt(table, 0, TPLID_MAX);
  if ((gm->type & GM_POSS)      /* for poss. graphical models */
  &&  (tab_opc(table, TAB_COND) != 0))
    return NAN;                 /* compute one point coverages */
  gm_clear(gm, 1);              /* clear the graphical model params. */
  if ((gm_estim(gm, table, 0) != 0)
  ||  isnan(gm_eval(gm, measure, params)))
    return NAN;                 /* estimate parameters and evaluate */
  gm_setnorm(gm, tplcnt);       /* set number of tuples for norm. */
  gmid.gm        = gm;          /* initialize the */
  gmid.table     = table;       /* induction variables */
  gmid.search    = search;
  gmid.measure   = measure;
  gmid.params[0] = params[0];
  gmid.params[1] = params[1];
  gmid.thresh    = (minimp  <= 0) ? 16*DBL_EPSILON : minimp;
  gmid.mrgimp    = (mrgimp  <= 0) ? 0 : mrgimp;
  gmid.lwise     = lwise;
  gmid.maxcon    = (maxcon  >= gm->attcnt) ? gm->attcnt-1 : maxcon;
  gmid.ptmax     = (maxtest <= 0) ? 1 : maxtest;
  gmid.log       = log;
  if (gmu_init(&gmid, gm->attcnt, (search == GM_TOPORD) ? 1 : 0) < 0)
    return NAN;                 /* create edge table (half or full) */
  err = (search == GM_NOLOOP)   /* induce a network structure */
      ? noloop(&gmid) : consel(&gmid);
  gmu_cleanup(&gmid);           /* clean up the induction variables */
  if (err) return NAN;          /* and check for an error */
  gm_clear(gm, 1);              /* clear the graphical model */
  if ((gm_estim(gm, table, 0) != 0)
  ||  isnan(gm_eval (gm, measure, params)))
    return NAN;                 /* rebuild the graphical model */
  if ((lwise >= 0) && isnan(gm_local(gm, lwise, mrgimp)))
    return NAN;                 /* redo local structure learning */
  return gm->eval;              /* return the evaluation result */
}  /* gm_consel() */
