/*----------------------------------------------------------------------
  File    : gramod5.c
  Contents: graphical model management (based on directed graphs)
            network induction with simulated annealing
  Author  : Christian Borgelt
  History : 2002.03.03 file created as gramod2.c
            2006.03.04 simulated annealing functions moved here
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#ifndef GM_EVAL
#define GM_EVAL
#endif
#ifndef GM_INDUCE
#define GM_INDUCE
#endif
#include "gramod.h"
#ifndef NODEID
#define NODEID      ATTID
#endif
#include "hyper.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- simulated annealing --- */
  TABLE  *table;                /* table to induce model from */
  TABLE  *test;                 /* table to test   model on */
  HTREE  *best;                 /* best hypertree found so far */
  HTREE  *curr;                 /* current hypertree */
  HTREE  *tmp;                  /* test hypertree */
} SIAN;                         /* (simulated annealing) */

/*----------------------------------------------------------------------
  Simulated Annealing
----------------------------------------------------------------------*/

static void cleanup (SIAN *sian)
{                               /* --- clean up induction variables */
  assert(sian);                 /* check the function argument */
  if (sian->table != sian->test)
    tab_delete(sian->table, 0); /* delete the estimation table */
  if (sian->best) ht_delete(sian->best);
  if (sian->curr) ht_delete(sian->curr);
  if (sian->tmp)  ht_delete(sian->tmp);
}  /* cleanup() */              /* delete the hypertrees */

/*--------------------------------------------------------------------*/

static int ht2gm (HTREE *ht, GRAMOD *gm)
{                               /* --- build graphical model */
  ATTID i, k, n;                /* loop variables */
  ATTID *nodes;                 /* to traverse the node identifiers */
  ATTID *marks;                 /* node markers */

  assert(ht && gm               /* check the function argument */
  &&    (ht_nodecnt(ht) == gm->attcnt));
  if (ht_reduce(ht) != 0)       /* do Graham reduction to determine */
    return -1;                  /* a construction sequence */
  for (i = gm->attcnt; --i >= 0; )
    gm_conrem(gm, i, -1);       /* remove all conditions */
  for (marks = gm->buf +(i = gm->attcnt); --i >= 0; )
    *--marks = 0;               /* initialize the node markers */
  if (gm->type == GM_PROB) {    /* --- if in probabilistic mode */
    for (n = 0; n < ht_hecnt(ht); n++) {
      nodes = ht_henodes(ht, n);/* traverse the construction sequence */
      for (i = ht_hesize(ht, n); --i >= 0; ) {
        if (marks[nodes[i]])    /* traverse all unprocessed */
          continue;             /* nodes of the hyperedge */
        for (k = ht_hesize(ht, n); --k >= 0; ) {
          if (!marks[nodes[k]]) /* traverse all processed */
            continue;           /* nodes of the hyperedge */
          if (gm_conadd(gm, nodes[i], nodes[k]) < 0) return -1;
        }                       /* add all processed nodes */
        marks[nodes[i]] = 1;    /* as conditions/parents and */
      }                         /* mark the node as processed */
    } }
  else {                        /* --- if in possibilistic mode */
    for (n = 0; n < ht_hecnt(ht); n++) {
      nodes = ht_henodes(ht, n);/* traverse the construction sequence */
      for (i = ht_hesize(ht, n); --i >= 0; )
        if (!marks[nodes[i]])   /* traverse the nodes and */
          break;                /* find an unprocessed node */
      assert(i >= 0);           /* check whether a node was found */
      for (k = ht_hesize(ht, n); --k >= 0; ) {
        marks[nodes[k]] = 1;    /* mark the nodes as processed, */
        if (k == i) continue;   /* but then skip the chosen node */
        if (gm_conadd(gm, nodes[i], nodes[k]) < 0) return -1;
      }                         /* add all other nodes */
    }                           /* as conditions/parents */
  }
  return 0;                     /* return 'ok' */
}  /* ht2gm() */

/*--------------------------------------------------------------------*/

double gm_sian (GRAMOD *gm, TABLE *table, int trials, ATTID maxsize,
                double keep, double frac, double ppwgt, int distuv,
                RANDFN randfn)
{                               /* --- simulated annealing */
  SIAN   sian;                  /* induction variables */
  int    step  = 0;             /* simulated annealing step */
  double worst =  INFINITY;     /* worth of worst hypertree found */
  double best  = -INFINITY;     /* worth of best hypertree found */
  double curr  = -INFINITY;     /* worth of current hypertree */
  double tmp;                   /* worth of test hypertree */
  double range;                 /* range of evaluation results */
  double fact = 1, mul;         /* factor for deviation */
  double res[3];                /* evaluation results */

  assert(gm && table            /* check the function arguments */
  &&    (trials > 0) && (maxsize >= 1) && (keep >= 0) && (keep <= 1)
  &&    (frac  >= 0) && (ppwgt >= 0));
  sian.table = sian.test = table;  /* note the data table and */
  sian.best  = sian.curr = NULL;   /* create an empty hypertree */
  sian.tmp   = ht_create(gm->attcnt);
  if (!sian.tmp) { cleanup(&sian); return NAN; }
  if (gm->type == GM_POSS) {    /* if to induce a poss. network */
    sian.table = tab_clone(table, 0);
    if (!sian.table || (tab_opc(sian.table, TAB_COND) != 0)) {
      cleanup(&sian); return NAN; }
  }                             /* compute one point coverages */
  gm_setnorm(gm, tab_getwgt(table, 0, TPLID_MAX));
  if (frac <= 0) frac = 0.001;  /* check and adapt the fraction */
  mul = pow(frac, 1.0/trials);  /* and compute the multiplier */
  while (1) {                   /* simulated annealing loop */
    if (((++step & 0x0f) == 0)) /* print a log message */
      fprintf(stderr, "%8d\b\b\b\b\b\b\b\b", step);
    if ((ht2gm(sian.tmp, gm) != 0)
    ||  (gm_estim(gm, sian.table, distuv) != 0)
    ||  (gm_evalx(gm, sian.test, res)     <  0)) {
      cleanup(&sian); return NAN; }
    tmp = ((gm->type & GM_POSS) ? -res[2] : res[0])
      - ppwgt *(double)gm_parsum(gm); /* evaluate graphical model */
    if (tmp < worst)            /* if it is worse than the worst, */
      worst = tmp;              /* note its evaluation */
    if (tmp > best) {           /* if it is better than the best */
      if (sian.best) ht_delete(sian.best);
      sian.best = ht_clone(sian.tmp);
      if (!sian.best) { cleanup(&sian); return NAN; }
      best = tmp;               /* copy the hypertree */
    }                           /* and its evaluation */
    range = ((step +1.0) /step) *(best -worst);
    if (range <= 0) range = 1;  /* compute estimate of eval. range */
    if ((tmp > curr)            /* alway select better result and */
    || ((tmp < curr)            /* select worse result randomly */
    &&  (randfn() *range *fact > curr -tmp))) {
      if (sian.curr) ht_delete(sian.curr);
      sian.curr = sian.tmp; sian.tmp = NULL;
      curr = tmp;               /* copy the hypertree */
    }                           /* and its evaluation */
    if (sian.tmp) { ht_delete(sian.tmp); sian.tmp = NULL; }
    if (step >= trials) break;  /* if all trials done, abort loop */
    sian.tmp = ht_modify(sian.curr, maxsize, keep, randfn);
    if (!sian.tmp) { cleanup(&sian); return NAN; }
    fact *= mul;                /* create modified hypertree and */
  }                             /* decrease the probability factor */
  if (sian.curr)  { ht_delete(sian.curr); sian.curr = NULL; }
  if (!sian.best) { cleanup(&sian); return NAN; }
  if ((ht2gm(sian.best, gm) != 0)
  ||  (gm_estim(gm, sian.table, distuv) != 0)
  ||  (gm_evalx(gm, sian.test, res)     <  0)) {
    cleanup(&sian); return NAN; }
  best = ((gm->type & GM_POSS) ? -res[2] : res[0])
    - ppwgt *(double)gm_parsum(gm);    /* reevaluate the best model */
  cleanup(&sian);               /* clean up the induction variables */
  return best;                  /* return the evaluation result */
}  /* gm_sian() */
