/*----------------------------------------------------------------------
  File    : graph.c
  Contents: Management of mixed graphs for graphical model construction
  Author  : Christian Borgelt
  History : 2005.12.22 file created
            2005.12.27 function gr_direct completed
            2006.01.02 dynamic adjacent node vectors introduced
            2006.01.03 functions gr_addundir, gr_remundir etc. added
            2006.01.24 function gr_common added (find common neighbors)
            2006.02.01 parameter mode added to function gr_adjsto
            2006.02.02 function gr_transto added (transitive neighbors)
            2006.03.02 some adjacent nodes functions added/redfined
            2006.03.03 parameter 'mode' added to function gr_common
            2006.03.09 extended functions added (use attribute set)
            2006.03.10 bug in function gr_direct fixed
            2006.03.11 function gr_dirpath added
            2006.12.18 function gr_undirect and gr_moralize added
            2007.01.28 gr_adjsto, gr_common generalized, gr_maxclq added
            2013.09.04 adapted to definition of NODEID
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#ifdef GR_PARSE
#include <string.h>
#endif
#include "graph.h"
#ifdef GR_EXTFN
#include "scanner.h"
#endif
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE        8        /* block size for node vectors */

/* --- error messages --- */
/* error codes 0 to -15 defined in scanner.h */
#define E_EDGE      (-16)       /* invalid edge type */

#ifdef GR_MAIN
/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_PARSE      (-2)       /* parse error */
#define E_SRC        (-3)       /* invalid source      node */
#define E_DST        (-4)       /* invalid destination node */
#define E_TOPORD     (-5)       /* no topological order found */
#define E_UNKOP      (-6)       /* unknown operation */
#endif

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef GR_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_EDGE    -16 */  "#invalid edge type %s",
};
#endif

#ifdef GR_MAIN
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_PARSE    -2 */  "parse error(s) on file %s",
  /* E_SRC      -3 */  "invalid source node %"NODEID_FMT,
  /* E_DST      -4 */  "invalid destination node %"NODEID_FMT,
  /* E_TOPORD   -5 */  "no topological order found",
  /* E_UNKOP    -6 */  "unknown operation %s",
  /*            -7 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef GR_MAIN
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan  = NULL;   /* scanner for the input file */
static GRAPH   *graph = NULL;   /* created graph */
static NODEID  *buf   = NULL;   /* buffer for node indices */
#endif

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

GRAPH* gr_create (NODEID nodecnt)
{                               /* --- create a graph */
  GRAPH  *graph;                /* created graph */
  NODE   *node;                 /* to traverse the nodes */
  NODEID size;                  /* size of adjacent nodes vector */

  assert(nodecnt > 0);          /* check the function arguments */
  graph = (GRAPH*)malloc(sizeof(GRAPH)
                       +(size_t)(nodecnt-1) *sizeof(NODE*)
                       +(size_t) nodecnt    *sizeof(NODEID));
  if (!graph) return NULL;      /* allocate the base structure */
  graph->nodecnt = nodecnt;     /* and initialize the fields */
  graph->edgecnt = graph->dircnt = 0;
  graph->topord  = NULL;        /* invalidate topological order */
  size = (nodecnt < BLKSIZE) ? nodecnt : BLKSIZE;
  while (--nodecnt >= 0) {      /* traverse the nodes */
    graph->nodes[nodecnt] =     /* allocate memory for each node */
    node = (NODE*)malloc(sizeof(NODE) +(size_t)size *sizeof(NODEID));
    node->deg  = node->indeg = node->outdeg = node->mark = 0;
    node->size = BLKSIZE;       /* create vector of adjacent nodes */
  }                             /* and initialize the node fields */
  graph->attset = NULL;         /* clear the attribute set variable */
  return graph;                 /* return the created graph */
}  /* gr_create */

/*--------------------------------------------------------------------*/

void gr_delete (GRAPH *graph)
{                               /* --- delete a graph */
  NODEID i;                     /* loop variable */

  assert(graph);                /* check the function arguments */
  for (i = 0; i < graph->nodecnt; i++)
    free(graph->nodes[i]);      /* delete all nodes */
  free(graph);                  /* delete the graph itself */
}  /* gr_delete */

/*--------------------------------------------------------------------*/
#ifdef GR_EXTFN

GRAPH* gr_createx (ATTSET *attset)
{                               /* --- create a graph */
  GRAPH *graph;                 /* created graph */

  assert(attset);               /* check the function argument */
  graph = gr_create(as_attcnt(attset));
  graph->attset = attset;       /* create graph, store attribute set, */
  return graph;                 /* and return the created graph */
}  /* gr_createx() */

/*--------------------------------------------------------------------*/

void gr_deletex (GRAPH *graph, int delas)
{                               /* --- delete a graph */
  if (delas) as_delete(graph->attset);
  gr_delete(graph);             /* delete graph and attribute set */
}  /* gr_deletex() */

#endif
/*--------------------------------------------------------------------*/

int gr_addedge (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- add an edge to the graph */
  NODEID i, o;                  /* loop variable, size/offset */
  NODE   *s, *d;                /* source and destination node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  d = graph->nodes[dst];        /* get the destination */
  s = graph->nodes[src];        /* and the source node */
  for (i = s->deg; --i >= 0; )  /* find dest. in source's neighbors */
    if (s->adjs[i] == dst) break;
  if (i < 0) {                  /* if the edge was not found, */
    if (s->deg >= s->size) {    /* if not enough space in source */
      o = s->size +((s->size > BLKSIZE) ? s->size >> 1 : BLKSIZE);
      if (o > graph->nodecnt) o = graph->nodecnt;
      s = (NODE*)realloc(s, sizeof(NODE) +(size_t)o *sizeof(NODEID));
      if (s == NULL) return -1; /* allocate a new, larger node */
      graph->nodes[src] = s; s->size = o;
    }                           /* replace the old node */
    if (d->deg >= d->size) {    /* if not enough space in destination */
      o = d->size +((d->size > BLKSIZE) ? d->size >> 1 : BLKSIZE);
      if (o > graph->nodecnt) o = graph->nodecnt;
      d = (NODE*)realloc(d, sizeof(NODE) +(size_t)o *sizeof(NODEID));
      if (d == NULL) return -1; /* allocate a new, larger node */
      graph->nodes[dst] = d; d->size = o;
    }                           /* replace the old node */
    o = d->deg -d->outdeg;      /* get index of first outgoing edge */
    d->adjs[d->deg++]   = d->adjs[d->indeg];
    d->adjs[d->indeg++] = d->adjs[o];
    d->adjs[o]          = src;  /* add source to the parents of */
    s->adjs[s->deg++]   = dst;  /* the destination and destination */
    s->outdeg++;                /* to the children of the source */
    graph->edgecnt++;           /* and update the degree counters */
    graph->dircnt++; }          /* as well as the edge counters */
  else {                        /* if a corresponding edge was found, */
    o = s->deg -s->outdeg -1;   /* get index of last pure parent */
    if (i > o) return -2;       /* check whether edge already exists */
    s->adjs[i] = s->adjs[o];    /* swap destination to a position */
    s->adjs[o] = dst;           /* next to the bidirectional edges */
    s->outdeg++;                /* and update the node's out-degree */
    for (i = d->deg; --i >= 0;) /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[d->indeg];    /* swap source to a position */
    d->adjs[d->indeg++] = src;  /* next to the bidirectional edges */
    graph->dircnt--;            /* and update the node's in-degree */
  }                             /* as well as the dir. edge counter */
  if (graph->topord && (s->pos > d->pos))
    graph->topord = NULL;       /* invalidate topological order */
  return 0;                     /* return 'adding successful' */
}  /* gr_addedge */

/*--------------------------------------------------------------------*/

static int connect (GRAPH *graph, NODEID src, NODEID dst, int aug)
{                               /* --- add an undirected edge */
  NODEID i, o;                  /* loop variable, size/offset */
  NODE   *s, *d;                /* source and destination node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  d = graph->nodes[dst];        /* get the destination */
  s = graph->nodes[src];        /* and the source node */
  for (i = s->deg; --i >= 0; )  /* find dest. in source's neighbors */
    if (s->adjs[i] == dst) break;
  if (i < 0) {                  /* if there is no corresponding edge */
    if (s->deg >= s->size) {    /* if not enough space in source */
      o = s->size +((s->size > BLKSIZE) ? s->size >> 1 : BLKSIZE);
      if (o > graph->nodecnt) o = graph->nodecnt;
      s = (NODE*)realloc(s, sizeof(NODE) +(size_t)o *sizeof(NODEID));
      if (s == NULL) return -1; /* allocate a new, larger node */
      graph->nodes[src] = s; s->size = o;
    }                           /* replace the old node */
    if (d->deg >= d->size) {    /* if not enough space in destination */
      o = d->size +((d->size > BLKSIZE) ? d->size >> 1 : BLKSIZE);
      if (o > graph->nodecnt) o = graph->nodecnt;
      d = (NODE*)realloc(d, sizeof(NODE) +(size_t)o *sizeof(NODEID));
      if (d == NULL) return -1; /* allocate a new, larger node */
      graph->nodes[dst] = d; d->size = o;
    }                           /* replace the old node */
    s->adjs[s->deg++]   = s->adjs[s->indeg];
    s->adjs[s->indeg++] = dst;  /* add bidirected edge to the source */
    s->outdeg++;                /* and update the node's degrees */
    d->adjs[d->deg++]   = d->adjs[d->indeg];
    d->adjs[d->indeg++] = src;  /* add bidirected edge to the dest. */
    d->outdeg++;                /* and update the node's degrees */
    graph->edgecnt++; }         /* as well as the edge counter */
  else if (i >= s->indeg) {     /* if corresp. outgoing edge exists */
    if (!aug) return -2;        /* if not to augment edges, abort */
    s->adjs[i] = s->adjs[s->indeg];
    s->adjs[s->indeg++] = dst;  /* swap dest. next to bidir. edges */
    o = d->deg -d->outdeg;      /* get number of pure parents */
    for (i = o; --i >= 0; )     /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[--o];  /* swap source to a position */
    d->adjs[o] = src;           /* next to the bidirectional edges */
    d->outdeg++;                /* and update the node's out-degree */
    graph->dircnt--; }          /* as well as the dir. edge counter */
  else {                        /* if corresp. incoming edge exists */
    if (!aug)  return -2;       /* if not to augment edges, abort */
    o = s->deg -s->outdeg -1;   /* get index of last pure parent */
    if (i > o) return -2;       /* check whether bidir. edge exists */
    s->adjs[i] = s->adjs[o];    /* swap destination to a position */
    s->adjs[o] = dst;           /* next to the bidirectional edges */
    s->outdeg++;                /* and update the node's out-degree */
    for (i = d->deg; --i >= 0;) /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[d->indeg];   /* swap source to a position */
    d->adjs[d->indeg++] = src;  /* next to the bidirectional edges */
    graph->dircnt--;            /* and update the node's in-degree */
  }                             /* as well as the dir. edge counter */
  if (graph->topord && (s->pos > d->pos))
    graph->topord = NULL;       /* invalidate topological order */
  return 0;                     /* return 'adding successful' */
}  /* connect() */

/*--------------------------------------------------------------------*/

int gr_addundir (GRAPH *graph, NODEID src, NODEID dst)
{ return connect(graph, src, dst, -1); }

/*--------------------------------------------------------------------*/

int gr_remedge (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- remove an edge from the graph */
  NODEID i, o;                  /* loop variables, offset */
  NODE   *s, *d;                /* source and destination node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  d = graph->nodes[dst];        /* get the destination */
  s = graph->nodes[src];        /* and the source node */
  for (i = s->deg; --i >= 0; )  /* find destination in source */
    if (s->adjs[i] == dst) break;
  if (i >= s->indeg) {          /* if the edge is purely outgoing */
    s->adjs[i] = s->adjs[--s->deg];
    s->outdeg--;                /* remove edge from the source */
    o = d->deg -d->outdeg;      /* get the number of pure parents */
    for (i = o; --i >= 0; )     /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[--o];  /* remove edge from the destination */
    d->adjs[o] = d->adjs[--d->indeg];
    d->adjs[d->indeg] = d->adjs[--d->deg];
    graph->edgecnt--;           /* update the node's degrees */
    graph->dircnt--; }          /* as well as the edge counters */
  else {                        /* if the edge is not purely outgoing */
    o = s->deg -s->outdeg;      /* get index of first outgoing edge */
    if (i < o) return -2;       /* if the edge was not found, abort */
    s->adjs[i] = s->adjs[o];    /* swap destination to a position */
    s->adjs[o] = dst;           /* next to the incoming edges */
    s->outdeg--;                /* and update the node's degrees */
    for (i = d->indeg; --i >= 0; )    /* find source in */
      if (d->adjs[i] == src) break;   /* dest.'s neighbors */
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[--d->indeg]; /* swap source to a position */
    d->adjs[d->indeg] = src;    /* next to the outgoing edges and */
    graph->dircnt++;            /* and update the node's out-degree */
  }                             /* as well as the dir. edge counter */
  return 0;                     /* return 'removal successful' */
}  /* gr_remedge */

/*--------------------------------------------------------------------*/

int gr_remundir (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- remove an undirected edge */
  NODEID i, o;                  /* loop variable, offset */
  NODE   *s, *d;                /* source and destination node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  d = graph->nodes[dst];        /* get the destination */
  s = graph->nodes[src];        /* and the source node */
  for (i = s->deg; --i >= 0; )  /* find destination in source */
    if (s->adjs[i] == dst) break;
  if (i < 0) return -2;         /* check whether edge exists */
  o = s->deg -s->outdeg;        /* get index of first outgoing edge */
  if (i < o) {                  /* if the edge is purely incoming */
    s->adjs[i] = s->adjs[--o];  /* remove edge from the source */
    s->adjs[o] = s->adjs[--s->indeg];
    s->adjs[s->indeg] = s->adjs[--s->deg];
    for (i = d->deg; --i >= 0;) /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[--d->deg];
    d->outdeg--;                /* remove edge from the destination */
    graph->edgecnt--;           /* update the node's degrees */
    graph->dircnt--; }          /* as well as the edge counters */
  else if (i >= s->indeg) {     /* if the edge is purely outgoing */
    s->adjs[i] = s->adjs[--s->deg];
    s->outdeg--;                /* remove edge from the source */
    o = d->deg -d->outdeg;      /* get the number of pure parents */
    for (i = o; --i >= 0; )     /* find source in dest.'s neighbors */
      if (d->adjs[i] == src) break;
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i] = d->adjs[--o];  /* remove edge from the destination */
    d->adjs[o] = d->adjs[--d->indeg];
    d->adjs[d->indeg] = d->adjs[--d->deg];
    graph->edgecnt--;           /* update the node's degrees */
    graph->dircnt--; }          /* as well as the edge counters */
  else {                        /* if the edge is bidirectional */
    s->adjs[i]        = s->adjs[--s->indeg];
    s->adjs[s->indeg] = s->adjs[--s->deg];
    s->outdeg--;                /* remove edge from the source */
    for (i = d->indeg; --i >= 0; )  /* find source in */
      if (d->adjs[i] == src) break; /* dest.'s neighbors */
    assert(i >= 0);             /* check consistency of the graph */
    d->adjs[i]        = d->adjs[--d->indeg];
    d->adjs[d->indeg] = d->adjs[--d->deg];
    d->outdeg--;                /* remove edge from the destination */
    graph->edgecnt--;           /* and update the node's degrees */
  }                             /* as well as the edge counter */
  return 0;                     /* return 'removal successful' */
}  /* gr_remundir() */

/*--------------------------------------------------------------------*/

int gr_edge (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- check existence of an edge */
  NODEID i;                     /* loop variables, offset */
  NODE   *s;                    /* source node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  s = graph->nodes[src];        /* get the source node and */
  for (i = s->deg; --i >= 0; )  /* find destination in source */
    if (s->adjs[i] == dst) break;
  if (i <  0)        return 0;  /* check the node index and */
  if (i >= s->indeg) return 1;  /* return the corresponding code */
  return (i < s->deg -s->outdeg) ? -1 : 2;
}  /* gr_edge() */

/*--------------------------------------------------------------------*/

static int dpath (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- check for a directed path */
  NODEID i, chd;                /* loop variable, child node id. */
  NODE   *node;                 /* current source node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  node = graph->nodes[src];     /* get the current source node */
  if (node->mark)               /* if node is marked, there is a */
    return (node->mark < 0);    /* or directed path or a dir. cycle */
  node->mark = 1;               /* mark the source node as visited */
  for (i = node->deg; --i >= node->indeg; ) {
    chd = node->adjs[i];        /* traverse all (pure) children */
    if ((chd == dst) || dpath(graph, chd, dst)) {
      node->mark = 0; return -1; }  /* if the dest. is a child or */
  }                             /* reachable from a (pure) child, */
  node->mark = 0;               /* there is a directed path, */
  return 0;                     /* otherwise there is not */
}  /* dpath() */

/*--------------------------------------------------------------------*/

int gr_dirpath (GRAPH *graph, NODEID src, NODEID dst)
{                               /* --- check for a directed path */
  NODEID i;                     /* loop variable */
  NODE   *node;                 /* current source node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  node = graph->nodes[src];     /* get the current source node */
  for (i = node->deg; --i >= node->indeg; )
    if (node->adjs[i] == dst)   /* if the destination is a child, */
      return +1;                /* there is a trivial directed path */
  for (i = graph->nodecnt; --i >= 0; )
    graph->nodes[i]->mark = 0;  /* clear all node markers and */
  node->mark = 1;               /* mark the source node as visited */
  for (i = node->deg; --i >= node->indeg; ) {
    if (dpath(graph, node->adjs[i], dst)) {
      node->mark = 0; return -1; }  /* if the destination node is */
  }                             /* reachable from a (pure) child, */
  node->mark = 0;               /* there is a directed path, */
  return 0;                     /* otherwise there is not */
}  /* gr_dirpath() */

/*--------------------------------------------------------------------*/

NODEID gr_inoutdeg (GRAPH *graph, NODEID id)
{                               /* --- get number of undirected edges */
  NODE *node = graph->nodes[id];
  return node->indeg +node->outdeg -node->deg;
}  /* gr_inoutdeg() */

/*--------------------------------------------------------------------*/

NODEID gr_parcnt (GRAPH *graph, NODEID id)
{                               /* --- get number of (pure) parents */
  NODE *node = graph->nodes[id];
  return node->deg -node->outdeg;
}  /* gr_parcnt() */

/*--------------------------------------------------------------------*/

NODEID gr_chdcnt (GRAPH *graph, NODEID id)
{                               /* --- get number of (pure) children */
  NODE *node = graph->nodes[id];
  return node->deg -node->indeg;
}  /* gr_chdcnt() */

/*--------------------------------------------------------------------*/

const NODEID* gr_outgoing (GRAPH *graph, NODEID id)
{                               /* --- get outgoing edges end nodes */
  NODE *node = graph->nodes[id];
  return node->adjs +(node->deg -node->outdeg);
}  /* gr_outgoing */

/*--------------------------------------------------------------------*/

const NODEID* gr_children (GRAPH *graph, NODEID id)
{                               /* --- get (pure) children of a node */
  NODE *node = graph->nodes[id];
  return node->adjs +node->indeg;
}  /* gr_children */

/*--------------------------------------------------------------------*/

NODEID gr_common (GRAPH *graph, NODEID src, NODEID dst, int mode,
                  NODEID *adjs)
{                               /* --- get common neighbors */
  NODEID i, k, e, n;            /* loop variables, node counter */
  NODE   *s, *d;                /* source and dest. node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  s = graph->nodes[src];        /* get the source node */
  for (i = 0; i < s->deg; i++)  /* unmark the sources neighbors */
    graph->nodes[s->adjs[i]]->mark =  0;
  d = graph->nodes[dst];        /* get the destination node and */
  d->mark = 0;                  /* unmark it (must not be included) */
  if (mode & GR_CHILDREN) {     /* if to get only pure children, */
    i = d->indeg; e = d->deg; } /* get range of pure children */
  else {                        /* if to get general neighbors */
    i = (mode & GR_NOPARENT) ? d->deg -d->outdeg : 0;
    e = (mode & GR_NOCHILD)  ? d->indeg : d->deg;
  }                             /* get range of qualifying neighbors */
  for ( ; i < e; i++)           /* mark the destination's neighbors */
    graph->nodes[d->adjs[i]]->mark = -1;
  if (mode & GR_CHILDREN) {     /* if to get only pure children, */
    i = s->indeg; e = s->deg; } /* get range of pure children */
  else {                        /* if to get general neighbors */
    i = (mode & GR_NOPARENT) ? s->deg -s->outdeg : 0;
    e = (mode & GR_NOCHILD)  ? s->indeg : s->deg;
  }                             /* get range of qualifying neighbors */
  for (n = 0; i < e; i++) {     /* traverse the source's neighbors */
    k = s->adjs[i];             /* and collect the marked neighbors */
    if (graph->nodes[k]->mark == 0) continue;
    if (adjs) adjs[n] = k;      /* if a neighbor array is given, */
    n++;                        /* also collect the common neighbors, */
  }                             /* otherwise only count them */
  return n;                     /* return the number of neighbors */
}  /* gr_common() */

/*--------------------------------------------------------------------*/

static void mark (GRAPH *graph, NODE *node)
{                               /* --- recursively mark nodes */
  NODEID i;                     /* loop variable */
  NODE   *adj;                  /* to traverse the adjacent nodes */

  node->mark = 1;               /* mark the node as visited */
  for (i = node->deg; --i >= 0; ) {
    adj = graph->nodes[node->adjs[i]];
    if (!adj->mark) mark(graph, adj);
  }                             /* traverse the adjacent nodes and */
}  /* mark() */                 /* recursively mark unmarked nodes */

/*--------------------------------------------------------------------*/

NODEID gr_adjsto (GRAPH *graph, NODEID src, NODEID dst, int mode,
                  NODEID *adjs)
{                               /* --- get adjacency path neighbors */
  NODEID i, k, e, n;            /* loop variables, node counter */
  NODE   *s, *d;                /* source and destination node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  for (i = 0; i < graph->nodecnt; i++)
    graph->nodes[i]->mark = 0;  /* clear all node markers */
  s = graph->nodes[src];        /* get the source node and */
  s->mark = 1;                  /* mark it (sentinel for recursion) */
  d = graph->nodes[dst];        /* get the destination node and */
  mark(graph, d);               /* recursively mark the nodes from it */
  d->mark = 0;                  /* unmark the destination node */
  i = (mode & GR_NOPARENT) ? s->deg -s->outdeg : 0;
  e = (mode & GR_NOCHILD)  ? s->indeg : s->deg;
  for (n = 0; i < e; i++) {     /* get the proper index range and */
    k = s->adjs[i];             /* traverse the source's neighbors */
    if (graph->nodes[k]->mark == 0) continue;
    if (adjs) adjs[n] = k;      /* if a neighbor array is given, */
    n++;                        /* also collect the marked neighbors, */
  }                             /* otherwise only count them */
  return n;                     /* return the number of neighbors */
}  /* gr_adjsto */

/*--------------------------------------------------------------------*/

static void markex (GRAPH *graph, NODE *node, NODEID mark)
{                               /* --- recursively mark nodes */
  NODEID i, m;                  /* loop variable, marker */
  NODE   *adj;                  /* to traverse the adjacent nodes */

  node->mark = (node->mark < 0) /* mark the node as visited, */
             ? -mark : mark;    /* but keep registration flag */
  for (i = 0; i < node->deg; i++) {
    adj = graph->nodes[node->adjs[i]];
    m   = (adj->mark < 0) ? -adj->mark : adj->mark;
    if (m < mark) markex(graph, adj, mark);
  }                             /* traverse the adjacent nodes and */
}  /* markex() */               /* recursively mark unmarked nodes */

/*--------------------------------------------------------------------*/

NODEID gr_transto (GRAPH *graph, NODEID src, NODEID dst, int mode,
                   NODEID *adjs)
{                               /* --- get transitive adj. neighbors */
  NODEID i, j, k, m, n;         /* loop variables, node counter */
  NODE   *s, *d, *p, *q;        /* source and dest. node, buffers */

  assert(graph && adjs          /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt)
  &&    (dst >= 0) && (dst < graph->nodecnt));
  for (i = 0; i < graph->nodecnt; i++)
    graph->nodes[i]->mark = 0;  /* clear all node markers */
  s = graph->nodes[src];        /* get the source node and */
  s->mark =  1;                 /* mark it (sentinel for recursion) */
  d = graph->nodes[dst];        /* get the destination node and */
  mark(graph, d);               /* mark nodes recursively from it */
  d->mark = -1;                 /* unmark the destination node */
  i = 0; j = s->deg;            /* get the default index range */
  if (mode & GR_NOPARENT) {     /* if to exclude parent nodes */
    for (k = s->deg -s->outdeg; i < k; )
      graph->nodes[s->adjs[i++]]->mark = -1;
  }                             /* remark all parent nodes */
  if (mode & GR_NOCHILD) {      /* if to exclude child nodes */
    for (k = s->indeg;          j > k; )
      graph->nodes[s->adjs[--j]]->mark = -1;
  }                             /* remark all child nodes */
  for (n = 0; i < j; i++) {     /* traverse the index range */
    k = s->adjs[i];             /* get index of source's neighbor */
    p = graph->nodes[k];        /* remark the node as registered */
    if (p->mark > 0) { p->mark = -1; adjs[n++] = k; }
  }                             /* collect the marked neighbors */
  m = 1;                        /* init. the marker value */
  for (i = n; --i >= 0; ) {     /* traverse source's neighbors again */
    p = graph->nodes[adjs[i]];  /* get the neighbor node and */
    s->mark = p->mark = -(++m); /* mark it (sentinel for recursion) */
    markex(graph, d, m);        /* mark neighbors of this neighbor */
    for (j = p->deg; --j >= 0; ) {
      k = p->adjs[j];           /* traverse the transitive neighbors */
      q = graph->nodes[k];      /* that bear an original marker */
      if (q->mark >= m) { q->mark = -m; adjs[n++] = k; }
    }                           /* collect the marked nodes */
  }                             /* and clear their marker */
  return n;                     /* return the number of neighbors */
}  /* gr_transto */

/*--------------------------------------------------------------------*/

const NODEID* gr_topord (GRAPH *graph)
{                               /* --- construct a topological order */
  NODEID i, k, n;               /* loop variables, buffer */
  NODE   *node;                 /* to traverse the nodes */

  assert(graph);                /* check the function argument */
  if (graph->dircnt < graph->edgecnt)
    return NULL;                /* all edges must be directed */
  if (graph->topord)            /* if a topological order exists, */
    return graph->topord;       /* simply return it */
  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse the nodes and */
    node->mark = node->outdeg;  /* copy the out-degree */
  }                             /* to the node marker */
  n = graph->nodecnt;           /* init. position in top. order */
  do {                          /* ordering loop */
    for (i = k = 0; i < graph->nodecnt; i++) {
      node = graph->nodes[i];   /* traverse the childless nodes */
      if (node->mark > 0) continue;
      node->mark =              /* mark the processed node and */
      node->pos  = --n;         /* set position in topological order */
      for (k = node->indeg; --k >= 0; )
        graph->nodes[node->adjs[k]]->mark--;
    }                           /* remove marked child from parents */
  } while ((n > 0) && (k < 0)); /* until all nodes are numbered */
  if (n > 0) return NULL;       /* check for success */
  graph->topord = (NODEID*)(graph->nodes +graph->nodecnt);
  for (i = graph->nodecnt; --i >= 0; )
    graph->topord[graph->nodes[i]->pos] = i;
  return graph->topord;         /* build and return the top. order */
}  /* gr_topord() */

/*--------------------------------------------------------------------*/

static void dir2src (GRAPH *graph, NODEID src, NODEID i)
{                               /* --- direct a bidirectional edge */
  NODEID k, o;                  /* loop variable, index, offset */
  NODEID dst;                   /* index of destination node */
  NODE   *s, *d;                /* source and dest. node of edge */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt));
  s   = graph->nodes[src];      /* get the source node */
  assert((i < s->indeg) && (i >= s->deg -s->outdeg));
  dst = s->adjs[i];             /* get the destination index, */
  d   = graph->nodes[dst];      /* the destination node, and */
  o   = d->deg -d->outdeg;      /* the index of first outgoing edge */
  for (k = d->indeg; --k >= o;) /* find source in destination */
    if (d->adjs[k] == src) break;
  d->adjs[k]        = d->adjs[--d->indeg];
  d->adjs[d->indeg] = src;
  s->adjs[i]        = s->adjs[o = s->deg -s->outdeg];
  s->adjs[o]        = dst;      /* swap source and destination */
  s->outdeg--;                  /* to positions adjacent to the */
  graph->dircnt++;              /* nodes of bidirectional edges and */
}  /* dir2src() */              /* increment the degree counters */

/*--------------------------------------------------------------------*/

static void dir2dst (GRAPH *graph, NODEID src, NODEID i)
{                               /* --- direct a bidirectional edge */
  NODEID k, o;                  /* loop variable, index, offset */
  NODEID dst;                   /* index of destination node */
  NODE   *s, *d;                /* source and dest. node of edge */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt));
  s   = graph->nodes[src];      /* get the source node */
  assert((i < s->indeg) && (i >= s->deg -s->outdeg));
  dst = s->adjs[i];             /* get the destination index, */
  d   = graph->nodes[dst];      /* the destination node, and */
  o   = d->deg -d->outdeg;      /* the index of first outgoing edge */
  for (k = d->indeg; --k >= o;) /* find source in destination */
    if (d->adjs[k] == src) break;
  s->adjs[i]        = s->adjs[--s->indeg];
  s->adjs[s->indeg] = dst;
  d->adjs[k]        = d->adjs[o];
  d->adjs[o]        = src;      /* swap source and destination */
  d->outdeg--;                  /* to positions adjacent to the */
  graph->dircnt++;              /* nodes of bidirectional edges and */
}  /* dir2dst() */              /* increment the degree counters */

/*--------------------------------------------------------------------*/

static void direct (GRAPH *graph, NODEID src)
{                               /* --- rec. direct undirected edges */
  NODEID i, d;                  /* loop variables, buffer */
  NODE   *node;                 /* the current node */

  assert(graph                  /* check the function arguments */
  &&    (src >= 0) && (src < graph->nodecnt));
  node = graph->nodes[src];     /* get the current node */
  if (node->mark) return;       /* if there is a dir. cycle, abort */
  node->mark = -1;              /* mark node for recursive processing */
  for (i = node->indeg; i < node->deg; i++)
    direct(graph, node->adjs[i]);   /* recursively directed the edges */
  while (1) {                   /* traverse the undirected edges */
    /* Note that in this loop the node degrees (deg, indeg, outdeg) */
    /* must be retrieved again, since they may have been changed    */
    /* in the recursive call carried out in the second case below.  */
    i = node->indeg;            /* get index of last undirected edge */
    if (i <= node->deg -node->outdeg)
      break;                    /* check for an empty index range */
    d = node->adjs[--i];        /* traverse the undirected edges */
    if ((graph->nodes[d]->mark) /* if adjacent node is an anchestor */
    ||  dpath(graph, d, src))   /* (w.r.t. current path or generally) */
      dir2src(graph, src, i);   /* direct edge towards current node */
    else {                      /* if the adj. node is no ancestor, */
      dir2dst(graph, src, i);   /* direct the edge towards the node */
      direct(graph, d);         /* recursively directed the edges */
    }                           /* away from the current node */
  }
  node->mark = 0;               /* unmark the node again */
}  /* direct() */

/*--------------------------------------------------------------------*/

NODEID gr_direct (GRAPH *graph, int full)
{                               /* --- direct undirected edges */
  NODEID i, k, n;               /* loop variables, buffer */
  NODE   *node;                 /* to traverse the nodes */

  assert(graph);                /* check the function argument */
  n = graph->dircnt;            /* note the number of directed edges */
  for (i = 0; i < graph->nodecnt; i++)
    graph->nodes[i]->mark = 0;  /* clear all node markers */

  /* --- avoid cycles and extend chains --- */
  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse the nodes of the graph */
    if ((node->deg >  node->outdeg)  /* if node has a pure parent */
    ||  (node->deg <= node->indeg))  /* or no pure outgoing edge, */
      continue;                 /* it is not eligible as a root */
    node->mark = -1;            /* mark the node as a root */
    for (k = node->deg; --k >= node->indeg; )
      direct(graph, node->adjs[k]);
    node->mark = 0;             /* recursively direct the edges, */
  }                             /* then unmark the node again */
  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse the nodes of the graph */
    k    = node->deg -node->outdeg;
    if ((k > 0) && (k < node->indeg))
      direct(graph, i);         /* for nodes with pure parents, */
  }                             /* recursively direct the edges */
  /* The above loop leads to newly oriented edges only */
  /* if there are directed cycles in the original graph. */
  if (!full)                    /* if not to direct fully, return */
    return graph->dircnt -n;    /* number of newly directed edges */

  /* --- direct remaining edges --- */
  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse the nodes of the graph */
    if (node->deg -node->outdeg < node->indeg)
      direct(graph, i);         /* for nodes with undirected edges */
  }                             /* recursively direct the edges */
  return graph->dircnt -n;      /* return number of directed edges */
}  /* gr_direct() */

/*--------------------------------------------------------------------*/

NODEID gr_undirect (GRAPH *graph)
{                               /* --- undirect directed edges */
  NODEID i, n = 0;              /* loop variable */
  NODE   *node;                 /* to traverse the nodes */

  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse all nodes */
    n += (node->deg -node->outdeg) -(node->deg -node->indeg);
    node->indeg = node->outdeg = node->deg;
  }                             /* make all incident edges undirected */
  return n;                     /* return number of undirected edges */
}  /* gr_undirect() */

/*--------------------------------------------------------------------*/

NODEID gr_moralize (GRAPH *graph)
{                               /* --- form moral graph */
  NODEID i, k1, k2, n = 0;      /* loop variables, counter */
  NODE   *node;                 /* to traverse the nodes */

  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse all nodes */
    for (k1 = node->deg -node->outdeg; --k1 >= 0; )
      for (k2 = k1; --k2 >= 0; )
        if (connect(graph, node->adjs[k1], node->adjs[k2], 0))
          n++;                  /* (try to) add an edge between */
  }                             /* all pairs of parent nodes */
  return n;                     /* return the number of added edges */
}  /* gr_moralize() */

/*--------------------------------------------------------------------*/

NODEID gr_maxclq (GRAPH *graph, NODEID *clique, NODEID cnt)
{                               /* --- get next maximal clique */
  NODEID i, k, n, old;          /* loop variables, old clique size */
  NODE   *node = NULL;          /* to traverse the nodes */

  if (cnt <= 0) {               /* if to initialize the search */
    for (i = 0; i < graph->nodecnt; i++)
      graph->nodes[i]->mark = 0;/* clear all node markers and */
    clique[0] = -1; cnt = 1;    /* set a dummy clique that will */
  }                             /* be extended into the first clique */
  while (--cnt >= 0) {          /* while there are nodes to remove */
    i = clique[old = cnt];      /* get the last node of the clique */
    if (i >= 0)                 /* and remove and unmark it */
      graph->nodes[i]->mark = 0;/* (try finding a clique without it) */
    while (++i < graph->nodecnt) {
      node = graph->nodes[i];   /* traverse the remaining nodes */
      for (n = 0, k = node->deg; --k >= 0; )
        if (graph->nodes[node->adjs[k]]->mark) n++;
      if (n != cnt) continue;   /* get number of marked neighbors and */
      clique[cnt++] = i;        /* if it is the current clique size, */
      node->mark    = 1;        /* the current clique can be extended */
    }                           /* so add the node and mark it */
    if (cnt <= old) continue;   /* if no ext. found, reduce further */
    for (i = clique[cnt-1]; --i >= 0; ) {
      node = graph->nodes[i];   /* traverse the preceding nodes */
      if (node->mark) continue; /* that are not in the clique */
      for (n = 0, k = node->deg; --k >= 0; )
        if (graph->nodes[node->adjs[k]]->mark) n++;
      if (n == cnt) break;      /* if the constructed clique can */
    }                           /* be extended, it is not maximal, */
    if (i < 0) return cnt;      /* otherwise, it is maximal, */
  }                             /* so return the number of nodes */
  return 0;                     /* return 'no further clique' */
}  /* gr_maxclq() */

/*--------------------------------------------------------------------*/

void gr_desc (GRAPH *graph, FILE *file, int mode, int maxlen)
{                               /* --- show a graph */
  int    f = 1;                 /* flag for separator output */
  NODEID i, k;                  /* loop variables */
  NODE   *node;                 /* to traverse the nodes */
  #ifdef GR_EXTFN               /* if to use extended functions */
  char   buf[4*AS_MAXLEN+4];    /* output buffer */
  #endif

  if (mode & GR_TITLE) {        /* if to print a title */
    i = k = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); while (--i >= 0) putc('-', file);
    fputs("\n  graph\n", file);
    while (--k >= 0) { putc('-', file); } fputs("*/\n", file);
  }
  fprintf(file, "graph(%"NODEID_FMT",%"NODEID_FMT") = {\n",
          graph->nodecnt, graph->edgecnt);
  for (i = 0; i < graph->nodecnt; i++) {
    node = graph->nodes[i];     /* traverse all nodes */
    for (k = node->deg -node->outdeg; k < node->indeg; k++) {
      if (node->adjs[k] <= i)   /* traverse the undirected edges */
        continue;               /* that lead "forward" */
      if (!f) fprintf(file, ",\n"); else f = 0;
      #ifdef GR_EXTFN           /* if to use extended functions */
      if (graph->attset         /* if to print node names */
      &&  ((mode & GR_IDS) == 0)) {
        fputs("  ",   file);    /* indent the output line */
        scn_format(buf,att_name(as_att(graph->attset, i)), 0);
        fputs(buf,    file);    /* print the source node name */
        fputs(" -- ", file);    /* and the edge description */
        scn_format(buf,att_name(as_att(graph->attset,node->adjs[k])),0);
        fputs(buf,    file); }  /* print the destination node name */
      else                      /* if to print node identifiers */
      #endif                    /* (or no extended functions) */
      fprintf(file, "  %"NODEID_FMT" -- %"NODEID_FMT, i, node->adjs[k]);
    }                           /* print edge description */
    for (; k < node->deg; k++){ /* traverse the outgoing edges */
      if (!f) fprintf(file, ",\n"); else f = 0;
      #ifdef GR_EXTFN           /* if to use extended functions */
      if (graph->attset         /* if to print node names */
      &&  ((mode & GR_IDS) == 0)) {
        fputs("  ",   file);    /* indent the output line */
        scn_format(buf,att_name(as_att(graph->attset, i)), 0);
        fputs(buf,    file);    /* print the source node name */
        fputs(" -> ", file);    /* and the edge description */
        scn_format(buf,att_name(as_att(graph->attset,node->adjs[k])),0);
        fputs(buf,    file); }  /* print the destination node name */
      else                      /* if to print node identifiers */
      #endif                    /* (or no extended functions) */
      fprintf(file, "  %"NODEID_FMT" -> %"NODEID_FMT, i, node->adjs[k]);
    }                           /* print edge description, */
  }                             /* but only one direction */
  if (!f) fprintf(file, "\n");  /* end the preceding line */
  fprintf(file, "};\n");        /* terminate the description */
}  /* gr_desc() */

/*--------------------------------------------------------------------*/
#ifdef GR_PARSE

static int parse (SCANNER *scan, GRAPH **pgr)
{                               /* --- parse a graph description */
  NODEID n, x;                  /* number of nodes, exchange buffer */
  NODEID src, dst;              /* ids of source and destination */
  int    t;                     /* buffer for token/type/result */

  assert(scan && pgr);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'graph' */
  ||  (strcmp(scn_value(scan), "graph") != 0))
    SCN_ERROR(scan, E_STREXP, "graph");
  SCN_NEXT(scan);               /* consume 'graph' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  n = (NODEID)strtol(scn_value(scan), NULL, 0);
  if (n < 0) SCN_ERROR(scan, E_NUMBER);
  *pgr = gr_create(n);          /* get the number of nodes */
  if (!*pgr) SCN_ERROR(scan, E_NOMEM);
  SCN_NEXT(scan);               /* consume the number of nodes */
  if (scn_token(scan) == ',') { /* if number of edges follows */
    SCN_NEXT(scan);             /* consume ',' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    SCN_NEXT(scan);             /* consume the number of edges */
  }
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  while (scn_token(scan) != '}') { /* while not at end of edge list */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    src = (NODEID)strtol(scn_value(scan), NULL, 0);
    if ((src < 0) || (src >= n)) SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume the source id */
    t = scn_token(scan);        /* get the edge descriptor */
    if      (t == T_DASH) t =  0;
    else if (t == T_RGT)  t =  1;
    else if (t == T_LFT)  t = -1;
    else SCN_ERROR(scan, E_EDGE, scn_value(scan));
    SCN_NEXT(scan);             /* consume the edge descriptor */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    dst = (NODEID)strtol(scn_value(scan), NULL, 0);
    if ((dst < 0) || (dst >= n)) SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume the dest. id */
    if (t < 0) { x = src; src = dst; dst = x; }
    if (t == 0) t = gr_addundir(*pgr, src, dst);
    else        t = gr_addedge (*pgr, src, dst);
    if (t < 0) SCN_ERROR(scan, E_NOMEM);
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

GRAPH* gr_parse (SCANNER *scan)
{                               /* --- parse a graph description */
  GRAPH *graph = NULL;          /* created graph */

  assert(scan);                 /* check the function argument */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parse(scan, &graph) != 0) {
    if (graph) gr_delete(graph);/* parse a graph description */
    return NULL;                /* if an error occurred, */
  }                             /* delete the graph and abort */
  return graph;                 /* return the created graph */
}  /* gr_parse() */

#endif
/*--------------------------------------------------------------------*/
#if defined GR_PARSE && defined GR_EXTFN

static int parsex (SCANNER *scan, GRAPH *graph)
{                               /* --- parse a graph description */
  NODEID src, dst, x;           /* ids of source and destination */
  int    t;                     /* buffer for token/type/result */

  assert(scan && graph);        /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'graph' */
  ||  (strcmp(scn_value(scan), "graph") != 0))
    SCN_ERROR(scan, E_STREXP, "graph");
  SCN_NEXT(scan);               /* consume 'graph' */
  if (scn_token(scan) == '(') { /* if number of nodes follows */
    SCN_NEXT(scan);             /* consume '(' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
     n = (int)strtol(scn_value(scan), NULL, 0);
    if (n != as_attcnt(graph->attset))
      SCN_ERROR(scan,E_NUMBER); /* get and check the number of nodes */
    SCN_NEXT(scan);             /* consume the number of nodes */
    if (scn_token(scan) == ','){/* if number of edges follows */
      SCN_NEXT(scan);           /* consume ',' */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      SCN_NEXT(scan);           /* consume the number of edges */
    }
    SCN_CHAR(scan, ')');        /* consume ')' */
  }
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  while (scn_token(scan) != '}') { /* while not at end of edge list */
    t = scn_token(scan);        /* check for an attribute name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
    src = as_attid(graph->attset, scn_value(scan));
    if (src < 0) SCN_ERROR(scan, E_UNKATT);
    SCN_NEXT(scan);             /* get and consume the source id */
    t = scn_token(scan);        /* get the edge descriptor */
    if      (t == T_DASH) t =  0;
    else if (t == T_RGT)  t =  1;
    else if (t == T_LFT)  t = -1;
    else XERROR(E_EDGE, scn_value(scan));
    SCN_NEXT(scan);             /* consume the edge descriptor */
    t = scn_token(scan);        /* check for an attribute name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
    dst = as_attid(graph->attset, scn_value(scan));
    if (dst < 0) SCN_ERROR(scan, E_UNKATT);
    SCN_NEXT(scan);             /* get and consume the dest. id */
    if (t <  0) { x = src; src = dst; dst = x; }
    if (t == 0) t = gr_addundir(graph, src, dst);
    else        t = gr_addedge (graph, src, dst);
    if (t < 0) SCN_ERROR(scan, E_NOMEM);
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parsex() */

/*--------------------------------------------------------------------*/

GRAPH* gr_parsex (SCANNER *scan, ATTSET *attset)
{                               /* --- parse a graph description */
  GRAPH *graph = NULL;          /* created graph */

  assert(scan && attset);       /* check the function argument */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parsex(scan, graph) != 0) {
    gr_delete(graph);           /* parse a graph description */
    return NULL;                /* if an error occurred, */
  }                             /* delete the graph and abort */
  return graph;                 /* return the created graph */
}  /* gr_parsex() */

#endif
/*--------------------------------------------------------------------*/
#ifdef GR_MAIN

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (buf)   free(buf);        \
  if (graph) gr_delete(graph); \
  if (scan)  scn_delete(scan, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function 1 for testing */
  NODEID i, n;                  /* loop variable, number of nodes */
  char   *fname;                /* name of the input file */
  NODEID src = 0, dst = 1;      /* source and destination index */

  /* --- print startup/usage message --- */
  if (argc < 3) {               /* if no arguments are given */
    printf("usage: %s file op [pars]\n", argv[0]);
    printf("file   file containing a graph description\n");
    printf("op     operation to perform on the graph\n");
    printf("       (common/adjsto/transto/direct/dirall"
                  "/topord/maxclqs)\n");
    printf("pars   parameters for the operation\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the function */

  /* --- read the graph --- */
  scan  = scn_create();         /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the domain file */
  fname = (argv[1][0] == '-') ? "" : argv[1];
  if (scn_open(scan, NULL, fname) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  graph = gr_parse(scan);       /* parse the graph */
  if (!graph) error(E_PARSE, scn_name(scan));
  n = gr_nodecnt(graph);        /* get the number of nodes */
  fprintf(stderr, "[%"NODEID_FMT" node(s)] done.\n", n);

  /* --- evaluate the operation --- */
  if (argc > 3) src = (NODEID)strtol(argv[3], NULL, 0);
  if (argc > 4) dst = (NODEID)strtol(argv[4], NULL, 0);
  if ((src < 0) || (src >= n)) error(E_SRC, src);
  if ((dst < 0) || (dst >= n)) error(E_DST, dst);
  if      (strcmp(argv[2], "common") == 0) {
    buf = (NODEID*)malloc((size_t)n *sizeof(NODEID));
    if (!buf) error(E_NOMEM);   /* get common neighbors */
    n = gr_common(graph, src, dst, 0, buf);
    printf("common(%"NODEID_FMT",%"NODEID_FMT"):\n", src, dst); }
  else if (strcmp(argv[2], "adjsto") == 0) {
    buf = (NODEID*)malloc((size_t)n *sizeof(NODEID));
    if (!buf) error(E_NOMEM);   /* get adjacency neighbors */
    n = gr_adjsto(graph, src, dst, 0, buf);
    printf("adjsto(%"NODEID_FMT",%"NODEID_FMT"):\n", src, dst); }
  else if (strcmp(argv[2], "transto") == 0) {
    buf = (NODEID*)malloc((size_t)n *sizeof(NODEID));
    if (!buf) error(E_NOMEM);   /* get transitive neighbors */
    n = gr_transto(graph, src, dst, 0, buf);
    printf("transto(%"NODEID_FMT",%"NODEID_FMT"):\n", src, dst); }
  else if (strcmp(argv[2], "direct") == 0) {
    fprintf(stderr, "directed %"NODEID_FMT" edge(s)\n",
            gr_direct(graph, 0)); }
  else if (strcmp(argv[2], "dirall") == 0) {
    fprintf(stderr, "directed %"NODEID_FMT" edge(s)\n",
            gr_direct(graph, 1)); }
  else if (strcmp(argv[2], "topord") == 0) {
    buf = (NODEID*)gr_topord(graph);
    if (!buf) error(E_TOPORD);  /* get topological order */
    printf("topological order:\n"); }
  else if (strcmp(argv[2], "maxclqs") == 0) {
    buf = (NODEID*)malloc((size_t)n *sizeof(NODEID));
    if (!buf) error(E_NOMEM);
    printf("maximal cliques:\n"); n = 0;
    do {                        /* traverse the maximal cliques */
      n = gr_maxclq(graph, buf, n);
      for (i = 0; i < n; i++)   /* print the nodes of the clique */
        printf("%"NODEID_FMT"%c", buf[i], (i < n-1) ? ' ' : '\n');
    } while (n > 0);            /* while there was another clique */
    free(buf); return 0; }      /* delete buffer and abort program */
  else error(E_UNKOP);          /* unknown operation */

  /* --- print the result --- */
  if (!buf)                     /* if the graph was modified, */
    gr_desc(graph, stdout, 0,0);/* describe the processed graph */
  else                          /* if there is a list of indices */
    for (i = 0; i < n; i++)     /* print the node indices */
      printf("%"NODEID_FMT"%c", buf[i], (i < n-1) ? ' ' : '\n');

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
/*--------------------------------------------------------------------*/
#if 0

int main (void)
{                               /* --- main function 2 for testing */
  GRAPH *graph = NULL;          /* test graph */

  graph = gr_create(8);         /* create a test graph */
  gr_addedge (graph, 0, 2);
  gr_addedge (graph, 1, 2);
  gr_addundir(graph, 2, 3);
  gr_addundir(graph, 2, 4);
  gr_addundir(graph, 4, 5);
  gr_addedge (graph, 3, 6);
  gr_addedge (graph, 5, 6);
  gr_addundir(graph, 6, 7);

  printf("initial graph\n");
  printf("%d node(s), %d edge(s), %d directed\n",
         gr_nodecnt(graph), gr_edgecnt(graph), gr_dircnt(graph));
  gr_desc(graph, stdout, 0, 0); printf("\n");

  printf("testing edge removal:\n");
  gr_remedge (graph, 0, 2);
  printf("directed   edge (0,2) removed\n");
  gr_addedge (graph, 0, 2);
  printf("directed   edge (0,2) readded\n");
  gr_remundir(graph, 4, 5);
  printf("undirected edge (4,5) removed\n");
  gr_addundir(graph, 4, 5);
  printf("undirected edge (4,5) readded\n");
  gr_remedge (graph, 4, 5);
  printf("directed   edge (4,5) removed\n");
  gr_remedge (graph, 5, 4);
  printf("directed   edge (5,4) removed\n");
  gr_addundir(graph, 4, 5);
  printf("undirected edge (4,5) readded\n");
  gr_remedge (graph, 5, 4);
  printf("directed   edge (5,4) removed\n");
  gr_remedge (graph, 4, 5);
  printf("directed   edge (4,5) removed\n");
  gr_addundir(graph, 4, 5);
  printf("undirected edge (4,5) readded\n");
  gr_remedge (graph, 4, 5);
  printf("directed   edge (4,5) removed\n");
  gr_remundir(graph, 4, 5);
  printf("undirected edge (4,5) removed\n");
  gr_addundir(graph, 4, 5);
  printf("undirected edge (4,5) readded\n");
  gr_remedge (graph, 5, 4);
  printf("directed   edge (5,4) removed\n");
  gr_remundir(graph, 4, 5);
  printf("undirected edge (4,5) removed\n");
  gr_addundir(graph, 4, 5);
  printf("undirected edge (4,5) readded\n\n");

  printf("processed graph\n");
  printf("%d node(s), %d edge(s), %d directed\n",
         gr_nodecnt(graph), gr_edgecnt(graph), gr_dircnt(graph));
  gr_desc(graph, stdout, 0, 0); printf("\n");
  gr_delete(graph);
  return 0;
}  /* main() */

#endif
