/*----------------------------------------------------------------------
  File    : hyper.h
  Contents: hypertree management for simulated annealing
  Author  : Christian Borgelt
  History : 1997.03.30 file created as sian.c
            1999.12.28 redesign completed
            2002.03.17 hypertree functions made a separate module
            2013.09.04 adapted to definition of NODEID
----------------------------------------------------------------------*/
#ifndef __HYPER__
#define __HYPER__

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef NODEID
#define NODEID      int         /* node identifier */
#endif

/*--------------------------------------------------------------------*/

#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

/*--------------------------------------------------------------------*/

#if   NODEID==int
#ifndef NODEID_MAX
#define NODEID_MAX  INT_MAX     /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#define NODEID_FMT  "d"         /* printf format code for int */
#endif
#ifndef nid_qsort
#define nid_qsort   int_qsort
#endif

#elif NODEID==long
#ifndef NODEID_MAX
#define NODEID_MAX  LONG_MAX    /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#define NODEID_FMT  "ld"        /* printf format code for long */
#endif
#ifndef nid_qsort
#define nid_qsort   lng_qsort
#endif

#elif NODEID==ptrdiff_t
#ifndef NODEID_MAX
#define NODEID_MAX  PTRDIFF_MAX /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#  ifdef _MSC_VER
#  define NODEID_FMT  "Id"      /* printf format code for ptrdiff_t */
#  else
#  define NODEID_FMT  "td"      /* printf format code for ptrdiff_t */
#  endif                        /* in case C99 is not supported */
#endif
#ifndef nid_qsort
#define nid_qsort   dif_qsort
#endif

#else
#error "NODEID must be either 'int', 'long' or 'ptrdiff_t'"
#endif

/*--------------------------------------------------------------------*/

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- hyperedge --- */
  NODEID size;                  /* number of connected nodes */
  NODEID nodes[1];              /* array of node identifiers */
} HEDGE;                        /* (hyperedge) */

typedef struct {                /* --- connected component --- */
  NODEID id;                    /* identifier of representative */
  NODEID cnt;                   /* number of represented nodes */
} HCOMP;                        /* (connected component) */

typedef struct {                /* --- hypertree --- */
  NODEID nodecnt;               /* number of nodes */
  NODEID *nodes;                /* buffer for node identifiers */
  HCOMP  *comps;                /* connected components */
  NODEID hecnt;                 /* number of hyperedges */
  int    flags;                 /* flags, e.g. HT_FULL */
  HEDGE  **buf;                 /* buffer for hyperedges */
  HEDGE  *hedges[1];            /* vector of hyperedges */
} HTREE;                        /* (hypertree) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern HTREE*  ht_create  (NODEID nodecnt);
extern void    ht_delete  (HTREE *ht);
extern HTREE*  ht_clone   (HTREE *ht);

extern NODEID  ht_nodecnt (HTREE *ht);
extern NODEID  ht_hecnt   (HTREE *ht);
extern int     ht_headd   (HTREE *ht, NODEID *nodes, NODEID size);
extern NODEID  ht_hesize  (HTREE *ht, NODEID index);
extern NODEID* ht_henodes (HTREE *ht, NODEID index);

extern int     ht_reduce  (HTREE *ht);
extern int     ht_fill    (HTREE *ht, NODEID maxsize, RANDFN randfn);
extern HTREE*  ht_modify  (HTREE *ht, NODEID maxsize, double keep,
                           RANDFN randfn);
#ifndef NDEBUG
extern void    ht_show    (HTREE *ht);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define ht_nodecnt(t)    ((t)->nodecnt)
#define ht_hecnt(t)      ((t)->hecnt)
#define ht_hesize(t,i)   ((t)->hedges[i]->size)
#define ht_henodes(t,i)  ((t)->hedges[i]->nodes)

#endif
