/*----------------------------------------------------------------------
  File    : ptree2.c
  Contents: probability/possibility tree management
            description and parsing functions
  Author  : Christian Borgelt
  History : 1995.10.22 file created as ptree.c
            1996.05.06 modified to sparse tree construction
            1996.11.27 type of counters changed to floating point
            1997.02.28 tree type (PT_PROB/PT_POSS) added, redesign
            1999.02.20 adapted to structure changes, assertions added
            1999.02.24 module based on attribute set management
            1999.04.05 network output changed to absolute frequencies
            2001.05.26 adapted to changed module scan
            2001.10.11 adapted to changed module scan
            2001.11.01 parse error recovery corrected/improved
            2002.01.20 major redesign completed
            2002.03.05 output of additional information added
            2002.07.08 bug in function linkout (~PT_OCCUR) fixed
            2006.02.16 reading a cond. indep. test description added
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "ptree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
#ifdef PT_PARSE
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */
#define E_DUPATT    (-18)       /* duplicate attribute */
#define E_MISATT    (-19)       /* missing attribute */
#define E_INVATT    (-20)       /* invalid attribute */
#define E_VALEXP    (-21)       /* attribute value expected */
#define E_UNKVAL    (-22)       /* unknown attribute value */
#define E_DUPVAL    (-23)       /* duplicate attribute value */
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- tree output information --- */
  FILE   *file;                 /* output file */
  int    mode;                  /* description mode, e.g. PT_FULL */
  double total;                 /* global total of counters */
  WEIGHT lcorr;                 /* Laplace correction */
  char   *base;                 /* string for base indentation */
  int    ind;                   /* indentation (in characters) */
  int    pos;                   /* position in output line */
  int    max;                   /* maximal output line length */
  char   buf[4*AS_MAXLEN+68];   /* output buffer */
} PTOUT;                        /* (tree output information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef PT_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
  /* E_DUPATT  -18 */  "#duplicate attribute '%s'",
  /* E_MISATT  -19 */  "#missing attribute '%s'",
  /* E_INVATT  -20 */  "#invalid attribute '%s'",
  /* E_VALEXP  -21 */  "#attribute value expected instead of '%s'",
  /* E_UNKVAL  -22 */  "#unknown attribute value '%s'",
  /* E_DUPVAL  -23 */  "#duplicate attribute value '%s'",
};
#endif

/*----------------------------------------------------------------------
  Description Functions
----------------------------------------------------------------------*/

static void indent (PTOUT *out, int newline)
{                               /* --- indent output line */
  int i;                        /* loop variable */

  assert(out);                  /* check the function argument */
  if (newline) {                /* terminate line */
    fputc('\n', out->file); fputs(out->base, out->file); out->pos = 0; }
  for (i = out->pos; i < out->ind; i++)
    fputc(' ',  out->file);     /* indent next line */
  if (out->ind > out->pos) out->pos = out->ind;
}  /* indent() */               /* note new position */

/*--------------------------------------------------------------------*/

static void distout (PTOUT *out, PTLVL *lvl, PTLEAF *leaf)
{                               /* -- print prob./poss. distribution */
  VALID  i;                     /* loop variable for values */
  int    len;                   /* length of a value name */
  double frq, norm;             /* frequency and normalization factor */

  assert(out && lvl);           /* check the function arguments */
  if (out->total >= 0)          /* for a possibility tree */
    norm = out->total;          /* use the global total */
  else {                        /* for a probability tree */
    norm = (double)lvl->cnt *out->lcorr;
    if (leaf) norm += leaf->total; /* get a node specific total and */
  }                             /* compute the normalization factor */
  norm = (norm > 0) ? 100/norm : 1;
  for (i = 0; i < lvl->cnt; i++) { /* traverse the values */
    if (i > 0) {                /* print a separator if necessary */
      fputc(',', out->file); out->pos++; }
    len  = (int)scn_format(out->buf, att_valname(lvl->att, i), 0);
    frq  = (leaf) ? leaf->cnts[i] : 0; /* format the value name */
    len += sprintf(out->buf +len, ": %.16g", frq);
    if (out->mode & PT_REL) {   /* add the absolute frequency */
      frq = (frq +out->lcorr) *norm;
      len += sprintf(out->buf +len, " (%.1f%%)", frq);
    }                           /* add the relative frequency */
    if ((out->pos +len > out->max -4)   /* if the current line */
    &&  (out->pos      > out->ind))     /* would get too long, */
      indent(out, 1);           /* start a new line and indent */
    else if (i > 0) {           /* otherwise only print a blank */
      fputc(' ', out->file ); out->pos++; }
    fputs(out->buf, out->file); /* print the value information */
    out->pos += len;            /* update the output position */
  }
  fputc(' ', out->file);        /* add a blank at the end and */
  out->pos++;                   /* update the output position */
}  /* distout() */

/*--------------------------------------------------------------------*/

static void linkout (PTOUT *out, PTLVL *lvl, PTNODE *node)
{                               /* --- print a link to another node */
  ATTID i;                      /* loop variable for levels */
  int   len;                    /* length of a value name */

  assert(out && lvl && node);   /* check the function arguments */
  fputs("->(", out->file);      /* start a link to another node */
  out->pos = out->ind += 3;     /* and update the output position */
  for (i = 0; node->parent; i++) {
    (--lvl)->link = node->index & ~PT_OCCUR;
    node = node->parent;        /* climb up in the tree and */
  }                             /* build a path to the linked node */
  for (len = -1; --i >= 0; lvl++) {  /* traverse the path */
    if (len >= 0) {             /* print a separator if necessary */
      fputc(',', out->file); out->pos++; }
    len = (int)scn_format(out->buf, att_valname(lvl->att,lvl->link), 0);
    if ((out->pos +len > out->max -4)   /* if the current line */
    &&  (out->pos      > out->ind))     /* would get too long, */
      indent(out, 1);           /* start a new line and indent */
    fputs(out->buf, out->file); /* print the value name and */
    out->pos += len;            /* update the output position */
  }
  fputc(')', out->file);        /* terminate the link and */
  out->pos += 1; out->ind -= 3; /* update the output position */
}  /* linkout() */

/*--------------------------------------------------------------------*/

static void treeout (PTOUT *out, PTLVL *lvl, PTNODE *node)
{                               /* -- recursive output of tree */
  VALID  i;                     /* loop variable for values */
  PTNODE *child;                /* to traverse the children */
  int    len;                   /* length of a value name */
  int    wd;                    /* width of attribute values */

  assert(out && lvl);           /* check the function arguments */
  fputs("{ ", out->file);       /* start a tree branch/distribution */
  out->pos = out->ind += 2;     /* and update the output position */
  if (lvl->index < 0) {         /* if the node is on the leaf level */
    if (node || (out->mode & PT_FULL))
      distout(out, lvl, (PTLEAF*)node); }
  else {                        /* if the node is an inner node */
    wd = att_valwd(lvl->att,1); /* get the value width for indenting */
    for (len = -1, i = 0; i < lvl->cnt; i++) {
      child = (node) ? node->children[i] : NULL;
      if (!(out->mode & PT_FULL)/* if not to show a full tree */
      &&  (!child               /* and the subtree is empty */
      ||  !(child->index & PT_OCCUR)))
        continue;               /* ignore the current child */
      if (len >= 0) {           /* print a separator if necessary */
        fputc(',', out->file); indent(out, 1); }
      len = (int)scn_format(out->buf, att_valname(lvl->att, i), 0);
      fputs(out->buf,out->file);/* print the name of the value */
      out->pos += len; out->ind += wd;
      indent(out, 0);           /* update output position and indent */
      fputc(':', out->file); out->pos++; out->ind++;
      if (child                 /* if the child reference is a link */
      && ((child->parent != node) || ((child->index & ~PT_OCCUR) != i)))
        linkout(out, lvl+1, child); /* print the link description */
      else                      /* if the child reference is genuine */
        treeout(out, lvl+1, child); /* print the subtree */
      out->ind -= wd +1;        /* recursively print the subtree */
    }                           /* and reset the indentation */
    if ((out->pos >= out->max)  /* if the line would get too long */
    || ((out->ind <= 0) && (out->pos >= out->max -1)))
      indent(out, 1);           /* start a new line */
  }
  fputc('}', out->file);        /* terminate the tree branch and */
  out->pos++; out->ind -= 2;    /* update the output position */
}  /* treeout() */

/*--------------------------------------------------------------------*/

int pt_desc (PTREE *pt, FILE *file, int mode, int maxlen)
{                               /* --- describe a prob./poss. tree */
  int    len, l;                /* loop variables for comments */
  ATTID  i;                     /* loop variable for levels */
  PTLVL  *lvl;                  /* to traverse the tree levels */
  PTNODE *root;                 /* the root node of the tree */
  PTOUT  out;                   /* tree output information */

  assert(pt && file);           /* check the function arguments */

  /* --- print header (as a comment) --- */
  if (mode & PT_TITLE) {        /* if the title flag is set */
    len = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, (pt->type == PT_PROB) ? "\n  probability tree\n"
                                        : "\n  possibility tree\n");
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  /* --- initialize --- */
  out.file  = file;             /* note the output file, the desc. */
  out.mode  = mode;             /* mode, and the output line length */
  out.max   = (maxlen <= 0) ? INT_MAX : maxlen;
  out.lcorr = pt->lcorr;        /* note the Laplace correction and */
  if (pt->lcorr > 0) out.mode |= PT_FULL;/* set the full tree flag */
  pt_occur(pt);                 /* compute totals and set occ. flags */
  out.total = (pt->type == PT_PROB)
            ? -1 : pt->total +out.lcorr +FLT_MIN;
  out.base  = (mode & PT_INDENT) ? "  " : "";

  /* --- print attribute list --- */
  fputs(out.base, file);        /* indent the line */
  fputs((pt->type == PT_PROB) ? "prob(" : "poss(", file);
  out.pos = out.ind = 5;        /* start the condition tree output */
  i   = pt->concnt;             /* get the number of conditions */
  lvl = pt->levels;             /* and the tree level descriptions */
  len = (int)scn_format(out.buf, att_name(lvl[i].att), 0);
  fputs(out.buf, file);         /* print the attribute name */
  out.pos = out.ind += len;     /* and the condition indicator */
  if (i > 0) { fputc('|', file); out.pos = ++out.ind; }
  for (len = -1; --i >= 0; lvl++) { /* traverse the conditions */
    if (len >= 0) {             /* print a separator if necessary */
      fputc(',', file); out.pos++; }
    len = (int)scn_format(out.buf, att_name(lvl->att), 0);
    if ((out.pos +len > out.max -3)     /* if the current line */
    &&  (out.pos      > out.ind))       /* would get too long, */
      indent(&out, 1);          /* start a new line and indent */
    fputs(out.buf, file);       /* print the name of the condition */
    out.pos += len;             /* compute the new output position */
  }
  fputs(") =", file);           /* terminate the attribute list */

  /* --- print the condition tree --- */
  root = pt->levels[0].list;    /* get the root node of the tree */
  if (!root || !(root->index & PT_OCCUR))
    fputs(" { }", file);        /* if the tree is empty, do nothing */
  else {                        /* if the tree is not empty, */
    fprintf(file, "\n%s", out.base);    /* start on the next line and */
    out.ind = out.pos = 0;      /* init. the position and indentation */
    treeout(&out, pt->levels, root);
  }                             /* recursively print the tree */
  fputs(";\n", file);           /* terminate the tree output */

  /* --- print additional information (as a comment) --- */
  if (mode & PT_INFO)           /* if the add. info. flag is set */
    fprintf(file, "%s/* evaluation: %g */\n", out.base, pt->eval);
  return (ferror(file)) ? -1 : 0;
}  /* pt_desc() */              /* return the writing result */

/*----------------------------------------------------------------------
  Parser Functions
----------------------------------------------------------------------*/
#ifdef PT_PARSE

static int distin (PTREE *pt, SCANNER *scan)
{                               /* --- read prob./poss. distribution */
  int    t;                     /* buffer for a token */
  VALID  i;                     /* value identifier */
  PTLVL  *lvl;                  /* to access the leaf level desc. */
  PTLEAF *leaf;                 /* to access the leaf node */
  WEIGHT *c;                    /* to traverse the counters */
  double f;                     /* buffer for a value frequency */

  assert(pt && scan);           /* check the function arguments */
  if (pt_setcnt(pt, 0, 0) != 0) /* create a leaf node and the path */
    SCN_ERROR(scan, E_NOMEM);   /* to it by setting a counter */
  lvl  = pt->levels+pt->pathlen;/* get the current level */
  leaf = (PTLEAF*)pt->curr;     /* and the leaf node */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
    *--c = -1;                  /* clear all counters in the leaf */
  while (1) {                   /* value/frequency read loop */
    t = scn_token(scan);        /* check for a value name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_VALEXP);
    i = att_valid(lvl->att, scn_value(scan));
    if (i < 0)                       SCN_ERROR(scan, E_UNKVAL);
    if (leaf->cnts[i] >= 0)          SCN_ERROR(scan, E_DUPVAL);
    SCN_NEXT(scan);             /* get and check the attribute value */
    SCN_CHAR(scan, ':');        /* consume ':' */
    if (scn_token(scan) != T_NUM)    SCN_ERROR(scan, E_NUMEXP);
    f = strtod(scn_value(scan), NULL); /* get and check frequency */
    if ((f < 0) || (f > FLT_MAX))    SCN_ERROR(scan, E_NUMBER);
    leaf->cnts[i] = (WEIGHT)f;  /* set the value frequency */
    SCN_NEXT(scan);             /* and consume it */
    if (scn_token(scan) == '('){/* if a relative frequency follows */
      SCN_NEXT(scan);           /* consume '(' */
      if (scn_token(scan) != T_NUM)  SCN_ERROR(scan, E_NUMEXP);
      f = atof(scn_value(scan));/* get and check the rel. frequency */
      if ((f < 0) || (f > 100))      SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* consume the rel. frequency */
      SCN_CHAR(scan, '%');      /* consume '%' */
      SCN_CHAR(scan, ')');      /* consume ')' */
    }
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at the end of the list, abort, */
  }                             /* otherwise consume ',' */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
    if (*--c < 0) *c = 0;       /* clear all unset counters */
  return 0;                     /* return 'ok' */
}  /* distin() */

/*--------------------------------------------------------------------*/

static int linkin (PTREE *pt, SCANNER *scan, VALID index)
{                               /* --- read a link to another node */
  int   t;                      /* buffer for a token */
  VALID i;                      /* value identifier */
  ATTID k;                      /* loop variable for conditions */
  PTLVL *lvl;                   /* to traverse the tree levels */

  assert(pt && scan);           /* check the function arguments */
  SCN_NEXT(scan);               /* consume '->' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  lvl = pt->levels;             /* traverse the tree levels */
  for (k = pt->pathlen +1; --k >= 0; ) {
    t = scn_token(scan);        /* check for a value name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_VALEXP);
    i = att_valid(lvl->att, scn_value(scan));
    if (i < 0)                       SCN_ERROR(scan, E_UNKVAL);
    lvl->link = i; lvl++;       /* get and store the identifier */
    SCN_NEXT(scan);             /* and consume the value name */
    if (k > 0) { SCN_CHAR(scan, ','); }
  }                             /* check for a separator if necessary */
  SCN_CHAR(scan, ')');          /* consume ')' */
  if (pt_link(pt, index, NULL) != 0) SCN_ERROR(scan, E_NOMEM);
  return 0;                     /* install the link and return 'ok' */
}  /* linkin() */

/*--------------------------------------------------------------------*/

static int treein (PTREE *pt, SCANNER *scan)
{                               /* --- read condition tree */
  int    t;                     /* function result, token buffer */
  VALID  i, k;                  /* value identifier, counter */
  PTLVL  *lvl;                  /* to access the current level */
  PTNODE *node;                 /* to access the child nodes */
  int    err;                   /* parse error status */

  assert(pt && scan);           /* check the function arguments */
  SCN_CHAR(scan, '{');          /* consume '{' */
  if (scn_token(scan) == '}') { /* if the (sub)tree is empty, */
    SCN_NEXT(scan); return 0; } /* consume '}' and abort */

  /* --- read a value distribution (a leaf) --- */
  if (pt_atleaf(pt)) {          /* if the current node is a leaf, */
    t = distin(pt, scan);       /* read a probability or */
    if (t) return t;            /* possibility distribution */
    SCN_CHAR(scan, '}'); return 0;
  }                             /* consume the closing '}' */

  /* --- read subtrees --- */
  lvl = pt->levels +pt->pathlen;/* get the current level and */
  err = 0; k = 0;               /* initialize the read counter */
  while (scn_token(scan) != '}') { /* attribute value read loop */
    if (k++ > 0) {              /* if this is not the first value, */
      SCN_CHAR(scan, ','); }    /* consume ',' (separator) */
    t = scn_token(scan);        /* check for a value name */
    if ((t != T_ID) && (t != T_NUM))      SCN_ERROR(scan, E_VALEXP);
    i = att_valid(lvl->att, scn_value(scan));
    if (i < 0)                            SCN_ERROR(scan, E_UNKVAL);
    node = (pt->curr) ? pt->curr->children[i] : NULL;
    if (node && (node->index & PT_OCCUR)) SCN_ERROR(scan, E_DUPVAL);
    SCN_NEXT(scan);             /* consume the attribute value */
    SCN_CHAR(scan, ':');        /* consume ':' */
    if (scn_token(scan) == T_RGT) {
      if (node) SCN_ERROR(scan, E_CHREXP, '{');
      t = linkin(pt, scan, i);  /* if an arrow '->' follows, */
      if (t) return t; }        /* read a link to another node */
    else {                      /* if a subtree follows */
      pt_down(pt, i);           /* go down in the tree and */
      t = treein(pt, scan);     /* read the subtree recursively */
      if (pt->curr)             /* mark the current node as read */
        pt->curr->index |= PT_OCCUR;
      pt_up(pt, PT_DESCENT);    /* go back up in the tree */
      if (t) {                  /* if an error occurred */
        if      (t < E_FWRITE) {
          if (scn_recover(scan, ';', '{', '}', 1) == T_EOF) return 1; }
        else if (t < 0) return t;
        err = 1;                /* try to recover on errors, */
      }                         /* but abort on fatal errors */
    }
  }
  SCN_NEXT(scan);               /* consume '}' */
  return err;                   /* return parse error status */
}  /* treein() */

/*--------------------------------------------------------------------*/

static ATTID consin (ATTSET *attset, SCANNER *scan, PTREE **pt)
{                               /* --- read list of conditions */
  int   t;                      /* buffer for a token */
  ATTID i, k, n = 0;            /* loop variables, number of conds. */
  ATT   *att;                   /* conditioned attribute */

  if (scn_token(scan) != '|')   /* if there are no conditions, */
    return 0;                   /* abort the function */
  do {                          /* attribute read loop */
    SCN_NEXT(scan);             /* consume '|' or ',' */
    t = scn_token(scan);        /* get and check the attribute id. */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
    i = as_attid(attset, scn_value(scan));
    if (i < 0) SCN_ERROR(scan, E_UNKATT);
    if (i == pt_attid(*pt))          SCN_ERROR(scan, E_DUPATT);
    for (k = pt_concnt(*pt); --k >= 0; )
      if (i == pt_conid(*pt, k))     SCN_ERROR(scan, E_DUPATT);
    att = as_att(attset, i);    /* add the attribute as a condition */
    if (pt_conadd(*pt, n, att) != 0) SCN_ERROR(scan, E_NOMEM);
    SCN_NEXT(scan);             /* consume the attribute name */
    n++;                        /* and count the condition */
  } while (scn_token(scan) == ',');
  return n;                     /* while end of list not reached */
}  /* consin() */               /* return number of conditions */

/*--------------------------------------------------------------------*/

static int parse (ATTSET *attset, SCANNER *scan, int type, PTREE **pt)
{                               /* --- read prob./poss. distributions */
  int   t = -1;                 /* tree type, token buffer */
  ATTID i, k;                   /* loop variables, buffers */
  ATT   *att;                   /* conditioned attribute */

  assert(attset && scan);       /* check the function arguments */
  if (scn_token(scan) == T_ID){ /* check for 'prob' or 'poss' */
    if     ((strcmp(scn_value(scan), "prob") == 0)
    ||      (strcmp(scn_value(scan), "P")    == 0)) t = PT_PROB;
    else if (strcmp(scn_value(scan), "poss") == 0)  t = PT_POSS;
  }                             /* determine the tree type */
  if ((t < 0)                   /* if no proper token found */
  ||  (!(type & PT_AUTO)        /* or not in automatic mode */
  &&  ((type ^ t) & PT_POSS)))  /* and wrong token found, abort */
    SCN_ERROR(scan, E_STREXP, (type == PT_PROB) ? "prob" : "poss");
  if (type & PT_AUTO) type = t; /* set the proper tree type */
  SCN_NEXT(scan);               /* consume 'prob' or 'poss' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* get and check the attribute id. */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0) SCN_ERROR(scan, E_UNKATT);
  att = as_att(attset, i);      /* get the conditioned attribute */
  k   = att_getmark(att);       /* and its read marker */
  if (k != 0) SCN_ERROR(scan, (k > 0) ? E_DUPATT : E_INVATT);
  SCN_NEXT(scan);               /* and consume the attribute name */
  *pt = pt_create(att, type);   /* create a prob./poss. tree */
  if (!*pt) SCN_ERROR(scan, E_NOMEM);  /* for this attribute */
  k = consin(attset,scan,pt);   /* read a possible list of conditions */
  if (k < 0) return (int)k;     /* and check for success */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (treein(*pt, scan) != 0)   /* read the prob./poss. tree */
    return -1;                  /* recursively */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

static int citest (ATTSET *attset, SCANNER *scan, int type, PTREE **pt)
{                               /* --- read cond. indep. test */
  int   t;                      /* token buffer, function result */
  ATTID i;                      /* attribute identifier */

  assert(attset && scan);       /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for cond. indep. test */
  &&  (strcmp(scn_value(scan), "I")     != 0)
  &&  (strcmp(scn_value(scan), "D")     != 0)
  &&  (strcmp(scn_value(scan), "CI")    != 0)
  &&  (strcmp(scn_value(scan), "cind")  != 0)
  &&  (strcmp(scn_value(scan), "indep") != 0)
  &&  (strcmp(scn_value(scan), "dep")   != 0))
    SCN_ERROR(scan, E_STREXP, "CI");
  SCN_NEXT(scan);               /* consume 'CI' (or alternative) */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for a name (1st attribute) */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0) SCN_ERROR(scan, E_UNKATT);
  SCN_NEXT(scan);               /* create a prob./poss. tree */
  *pt = pt_create(as_att(attset, i), type & PT_POSS);
  if (!*pt) SCN_ERROR(scan, E_NOMEM);
  SCN_CHAR(scan, ',');          /* consume ',' */
  t = scn_token(scan);          /* check for a name (2nd attribute) */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0) SCN_ERROR(scan, E_UNKATT);
  if (i == pt_attid(*pt))         SCN_ERROR(scan, E_DUPATT);
  SCN_NEXT(scan);               /* consume the attribute name */
  if (pt_conadd(*pt, 0, as_att(attset, i)) < 0)
    SCN_ERROR(scan, E_NOMEM);   /* add end att. as first condition */
  i = consin(attset, scan, pt); /* read a possible list of conditions */
  if (i < 0) return (int)i;     /* and check for success */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* citest() */

/*--------------------------------------------------------------------*/

PTREE* pt_parse (ATTSET *attset, SCANNER *scan, int type)
{                               /* --- parse a prob./poss. tree */
  int   r;                      /* result of parsing */
  PTREE *pt = NULL;             /* created prob./poss. tree */

  assert(attset && scan);       /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  r = (type == PT_CITEST)       /* parse a prob./poss. tree */
    ? citest(attset, scan, type, &pt)
    : parse (attset, scan, type, &pt);
  if (r != 0) { if (pt) pt_delete(pt, 0); return NULL; }
  pt->parcnt =  0;              /* number of parameters is invalid */
  pt->occur  =  0;              /* occurrence flags are invalid */
  pt->total  = -1;              /* counter totals are invalid */
  if (type != PT_CITEST)        /* if not cond. independence test, */
    pt_total(pt);               /* compute the totals */
  return pt;                    /* return the created tree */
}  /* pt_parse() */

#endif
